from unittest import skip

import numpy as np
import torch

from pydtnn.activations.arctanh import Arctanh
from pydtnn.activations.leaky_relu import LeakyRelu
from pydtnn.activations.log import Log
from pydtnn.activations.relu import Relu
from pydtnn.activations.relu6 import Relu6
from pydtnn.activations.sigmoid import Sigmoid
from pydtnn.activations.softmax import Softmax
from pydtnn.activations.tanh import Tanh
from pydtnn.layers.adaptive_average_pool_2d import AdaptiveAveragePool2D
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.layers.max_pool_2d import MaxPool2D
from pydtnn.model import Model
from pydtnn.utils import random
from pydtnn.utils.tensor import TensorFormat, format_reshape, format_transpose
from pydtnn.tests.abstract.common import Params, TestCase, verbose_test
from pydtnn.utils.constants import Parameters


#from torch.testing._internal.common_utils import numpy_to_torch_dtype_dict
numpy_to_torch_dtype_dict = {
    np.bool_      : torch.bool,
    np.uint8      : torch.uint8,
    np.uint16     : torch.uint16,
    np.uint32     : torch.uint32,
    np.uint64     : torch.uint64,
    np.int8       : torch.int8,
    np.int16      : torch.int16,
    np.int32      : torch.int32,
    np.int64      : torch.int64,
    np.float16    : torch.float16,
    np.float32    : torch.float32,
    np.float64    : torch.float64,
    np.complex64  : torch.complex64,
    np.complex128 : torch.complex128
}

# setting random seed
SEED = 1234
random.seed(SEED)
# ---------
# ===============
# Constant values
# ===============

N = 64
C = 3
H = 524
W = 524

ADAPTIVE_AVG_POOL_OUTPUT_SIZE = (3, 3)

AVG_POOL_SHAPE = (3, 3)
AVG_POOL_PADDING = 0
AVG_POOL_STRIDE = 1

BATCH_NORMALIZATION_GAMMA = 1
BATCH_NORMALIZATION_BETA = 0
BATCH_NORMALIZATION_EPSILON = 1e-5
BATCH_NORMALIZATION_MOMENTUM_PYDTNN = 0.9
BATCH_NORMALIZATION_MOMENTUM_TORCH = BATCH_NORMALIZATION_MOMENTUM_PYDTNN
BATCH_NORMALIZATION_NUM_FEATURES = C

CONV2D_IN_C_TORCH = C
CONV2D_N_FILTERS = 5
CONV2D_FILTER_SHAPE = (4, 4)
CONV2D_PADDING = 0
CONV2D_STRIDE = 1
CONV2D_DILATION = 1

FC_OUPUT_SHAPE = (4, )
LINEAR_OUTPUT = FC_OUPUT_SHAPE[0]

MAX_POOL_SHAPE = (2, 2)
MAX_POOL_PADDING = 0
MAX_POOL_STRIDE = 1
MAX_POOL_DILATION = 1


GRAD_EQUIVALENCES: dict[str, str] = {
    Parameters.WEIGHTS: "weight",
    Parameters.BIASES: "bias",
    # Parameters.RUNNING_MEAN: "running_mean", # Not PyDTNN's grading var.
    # Parameters.RUNNING_VAR: "running_var", # Not PyDTNN's grading var.
    # Parameters.BETA: "", # Not in PyTorch
    # Parameters.GAMMA: "", # Not in PyTorch
}
# ==============

# ==============
# PyTorch models
# ==============

class TorchArcTanH(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.arc_tanh = torch.atanh

    def forward(self, x):
        x = self.arc_tanh(x)
        return x
# -------------


class TorchAdditionBlock(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.block1 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=CONV2D_IN_C_TORCH, out_channels=CONV2D_N_FILTERS, kernel_size=CONV2D_FILTER_SHAPE, padding=CONV2D_PADDING,
                            stride=CONV2D_STRIDE, dilation=CONV2D_DILATION),
            torch.nn.BatchNorm2d(CONV2D_N_FILTERS, eps=BATCH_NORMALIZATION_EPSILON, momentum=BATCH_NORMALIZATION_MOMENTUM_TORCH)
        )
        self.block2 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=CONV2D_IN_C_TORCH, out_channels=CONV2D_N_FILTERS, kernel_size=CONV2D_FILTER_SHAPE, padding=CONV2D_PADDING,
                            stride=CONV2D_STRIDE, dilation=CONV2D_DILATION)
        )

    def forward(self, x):
        x1 = self.block1(x)
        x2 = self.block2(x)
        x = x1 + x2
        return x
# -------------


class TorchConcatenationBlock(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.block1 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=CONV2D_IN_C_TORCH, out_channels=CONV2D_N_FILTERS, kernel_size=CONV2D_FILTER_SHAPE, padding=CONV2D_PADDING,
                            stride=CONV2D_STRIDE, dilation=CONV2D_DILATION),
            torch.nn.BatchNorm2d(CONV2D_N_FILTERS, eps=BATCH_NORMALIZATION_EPSILON, momentum=BATCH_NORMALIZATION_MOMENTUM_TORCH)
        )
        self.block2 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=CONV2D_IN_C_TORCH, out_channels=CONV2D_N_FILTERS, kernel_size=CONV2D_FILTER_SHAPE, padding=CONV2D_PADDING,
                            stride=CONV2D_STRIDE, dilation=CONV2D_DILATION)
        )

    def forward(self, x):
        x1 = self.block1(x)
        x2 = self.block2(x)
        x = torch.cat([x1, x2], dim=1)
        return x
# -------------
# ====================

class D:
    def __init__(self, b=N, c=C, h=H, w=W):
        self.b = b  # Batch size
        self.c = c  # Channels per layer
        self.h = h  # Layers height
        self.w = w  # Layers width
# ---

class ParamsLayerPytorch(Params):
    def __init__(self, d = D()) -> None:
        super().__init__()
        self.batch_size = d.b
        self.conv_variant = "i2c"
        self.tensor_format = TensorFormat.NCHW.upper()
        self.shape = format_reshape((C, H, W), "CHW", self.tensor_format[1:])
        self.model_name = None
        self.evaluate_only = True
        self.parallel = "sequential"
        self.loss_func = "categorical_cross_entropy"
        self.enable_gpu = False
        self.omm = None
        self.dtype = np.dtype(np.float32)
        self.tracing = False
        self.tracer_output = ""
        torch.set_default_dtype(numpy_to_torch_dtype_dict[self.dtype.type])
        self.dtype = np.dtype(self.dtype)
    
    def asdict(self):
        return self.__dict__
# ----

class LayerPyTorchTestCase(TestCase):

    params = ParamsLayerPytorch()

    def setUp(self) -> None:
        super().setUp()
        torch.manual_seed(0)
    # -----

    # ======================
    # Initialization methods
    # ======================

    @staticmethod
    def get_test_data(no_zeros=False, normalize=True, positives_and_negatives=True, 
                      shape_with_elements = (params.batch_size, *params.shape), dtype = params.dtype) -> np.ndarray:
        num_elems = np.prod(shape_with_elements) // 4

        x_1 = np.arange(num_elems)
        x_2 = np.arange(num_elems) * -1

        if no_zeros:
            x_1 += 1
            x_2 -= 1

        x_1_1 = np.where(x_1 % 2 == 0, x_1, x_1 / 3)
        x_1_2 = np.where(x_1 % 2 != 0, x_1, x_1)

        x_2_1 = np.where(x_1 % 2 == 0, x_2, x_2 / 3)
        x_2_2 = np.where(x_1 % 2 != 0, x_2, x_2)

        # NOTE: seems that PyTorch doesn't like too much np.float64
        x = np.stack([x_1_1, x_1_2, x_2_1, x_2_2], axis=0, dtype=dtype).reshape(shape_with_elements)
        random.shuffle(x)

        if normalize:
            min_x = np.min(x)
            x = (x - min_x) / (np.max(x) - min_x)
            if positives_and_negatives:
                x -= 0.5

        return np.asarray(x, dtype=dtype, order="C", copy=True)
    # ---------

    @staticmethod
    def initialize_pydtnn_model(list_layers: list[LayerBase], params=params) -> Model:
        model = Model(**params.asdict())
        model.add(Input(params.shape))
        model.add_layers(list_layers)
        model.mode = Model.Mode.TRAIN
        model._initialize()
        return model
    # ---------

    def _copy_grad_vars(self, grad: np.ndarray, grad_var: str, torch_layer: torch.nn.Module) -> None:
        if grad is not None:
            torch_grad_var = GRAD_EQUIVALENCES[grad_var]
            torch_grad = getattr(torch_layer, torch_grad_var)
            torch_grad.copy_(torch.from_numpy(grad.copy()).to(torch.device("cpu")).float())
    # ---

    def copy_grad_vars(self, pydtnn_model: Model, torch_model: torch.nn.Module) -> None:
        layers = pydtnn_model.get_all_layers().copy()

        if isinstance(layers[0], Input):
            layers.pop(0)

        torch_layers = [module for module in torch_model.modules() if not isinstance(module, torch.nn.Sequential)]
        # print(f"{layers=} {len(layers)=} || {len(torch_layers)=} {torch_layers=}")

        with torch.no_grad():
            for i in range(len(layers)):
                layer = layers[i]
                torch_layer = torch_layers[i]
                match layer:
                    case BatchNormalization():
                        running_mean = layer.running_mean
                        running_var = layer.running_var
                        if running_mean is not None:
                            torch_layer.running_mean.copy_(torch.from_numpy(running_mean.copy()).to(torch.device("cpu")).float())
                        if running_var is not None:
                            torch_layer.running_var.copy_(torch.from_numpy(running_var.copy()).to(torch.device("cpu")).float())
                    case FC():
                        for grad_var in layer.grad_vars.keys():
                            grad: np.ndarray = getattr(layer, grad_var)
                            grad = grad if grad_var != Parameters.WEIGHTS else grad.T
                            self._copy_grad_vars(grad, grad_var, torch_layer)
                    case Conv2D():
                        for grad_var in layer.grad_vars.keys():
                            grad: np.ndarray = getattr(layer, grad_var)
                            if grad_var == "weights" and grad is not None:
                                grad = format_transpose(grad, {TensorFormat.NHWC: "ihwo", TensorFormat.NCHW: "oihw"}[pydtnn_model.tensor_format], "oihw")
                            self._copy_grad_vars(grad, grad_var, torch_layer)
                    case _:
                        for grad_var in layer.grad_vars.keys():
                            grad: np.ndarray = getattr(layer, grad_var)
                            self._copy_grad_vars(grad, grad_var, torch_layer)
    # ----

    def do_test(self, _x: np.ndarray, pydtnn_model: Model, torch_model: torch.nn.Module, name_test: str, rtol=1e-6, atol=1e-6) -> None:
        self.copy_grad_vars(pydtnn_model, torch_model)

        num_elems = (len("Testing: ") + len(name_test))
        if verbose_test():
            print(f"\n\n{'=' * num_elems}\nTesting: {name_test}\n{'=' * num_elems}")

        x = np.copy(_x)

        x = x.astype(dtype=self.params.dtype, order="C", copy=None)

        for layer in pydtnn_model.layers:
            x: np.ndarray = layer.forward(x)
        x_pydtnn = x
        x_pydtnn = format_transpose(x, self.params.tensor_format.upper(), TensorFormat.NCHW.upper())

        x = torch.from_numpy(_x.reshape((N, C, H, W), copy=False)).to(torch.device("cpu")).float()
        x_torch: torch.Tensor = torch_model(x)
        x_torch = np.asarray(x_torch.cpu().detach().numpy(), dtype=pydtnn_model.dtype, order="C", copy=None)

        if verbose_test():
            print(f"[{rtol=}, {atol=}]\n{x_pydtnn.max()=}\n{x_torch.max()=}\n{x_pydtnn.min()=}\n{x_torch.min()=}\n{x_pydtnn.std()=}\n{x_torch.std()=}\n{x_pydtnn.mean()=}\n{x_torch.mean()=}")

        diff = x_pydtnn - x_torch
        if verbose_test():
            print(f"{diff.max()=}\n{diff.min()=}\n{diff.std()=}\n{diff.mean()=}")

        # if not (diff < rtol).all():
        #    print(f"x_pydtnn:\n{x_pydtnn}")
        #    print(f"x_torch:\n{x_torch}")
        #    print(f"diff:\n{diff}")

        # self.assertTrue((diff < rtol).all()), f"Not all values are below the rtol. Max. difference: \"{diff.max()}\". Std. deviation: \"{diff.std()}\". Min. difference: {diff.min()}."
        self.assertTrue(np.allclose(x_pydtnn, x_torch, rtol=rtol, atol=atol))
    # ---------
    # ====================

    # ====================
    # Unitary Test methods
    # ====================

    def test_AdaptiveAveragePool2D(self):
        pydtnn_layers = [AdaptiveAveragePool2D(output_shape=ADAPTIVE_AVG_POOL_OUTPUT_SIZE)]
        torch_model = torch.nn.AdaptiveAvgPool2d(output_size=ADAPTIVE_AVG_POOL_OUTPUT_SIZE)
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()

        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="AdaptiveAveragePool2D", rtol=1e-4, atol=1e-3)
    # ---------

    def test_AveragePool2D(self):
        pydtnn_layers = [AveragePool2D(pool_shape=AVG_POOL_SHAPE, padding=AVG_POOL_PADDING, stride=AVG_POOL_STRIDE)]
        torch_model = torch.nn.AvgPool2d(kernel_size=AVG_POOL_SHAPE, padding=AVG_POOL_PADDING, stride=AVG_POOL_STRIDE)
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="AveragePool2D")
    # ---------

    def test_BatchNormalization(self):
        pydtnn_layers = [BatchNormalization(gamma=BATCH_NORMALIZATION_GAMMA, beta=BATCH_NORMALIZATION_BETA, epsilon=BATCH_NORMALIZATION_EPSILON,
                                            momentum=BATCH_NORMALIZATION_MOMENTUM_PYDTNN)]
        torch_model = torch.nn.BatchNorm2d(BATCH_NORMALIZATION_NUM_FEATURES, eps=BATCH_NORMALIZATION_EPSILON, momentum=BATCH_NORMALIZATION_MOMENTUM_TORCH, affine=False)
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)

        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="BatchNormalization", rtol=1e0, atol=1e0)
    # ---------

    def test_Conv2D(self):
        pydtnn_layers = [Conv2D(grouping=Conv2D.Grouping.STANDARD, nfilters=CONV2D_N_FILTERS, filter_shape=CONV2D_FILTER_SHAPE, padding=CONV2D_PADDING, stride=CONV2D_STRIDE, dilation=CONV2D_DILATION)]
        torch_model = torch.nn.Conv2d(in_channels=CONV2D_IN_C_TORCH, out_channels=CONV2D_N_FILTERS, kernel_size=CONV2D_FILTER_SHAPE,
                                      padding=CONV2D_PADDING, stride=CONV2D_STRIDE, dilation=CONV2D_DILATION)
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Conv2D")
    # ---------

    @skip(reason="Dropout makes a random mask, then it can not be compared due both PyTorch and PyDTNN create different masks.")
    def test_Dropout(self):
        pydtnn_layers = [Dropout()]
        torch_model = torch.nn.Dropout()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Dropout")
    # ---------

    def test_Flatten(self):
        pydtnn_layers = [Flatten()]
        torch_model = torch.nn.Flatten()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Flatten")
    # ---------

    def test_FC(self):
        pydtnn_layers = [Flatten(), FC(shape=FC_OUPUT_SHAPE)]
        torch_model = torch.nn.Sequential(torch.nn.Flatten(), torch.nn.Linear(in_features=np.prod(self.params.shape), out_features=LINEAR_OUTPUT))
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="FC", rtol=1e-5, atol=1e-5)
    # ---------

    def test_MaxPool2D(self):
        pydtnn_layers = [MaxPool2D(pool_shape=MAX_POOL_SHAPE, padding=MAX_POOL_PADDING, stride=MAX_POOL_STRIDE, dilation=MAX_POOL_DILATION)]
        torch_model = torch.nn.MaxPool2d(kernel_size=MAX_POOL_SHAPE, padding=MAX_POOL_PADDING, stride=MAX_POOL_STRIDE, dilation=MAX_POOL_DILATION)
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="MaxPool2D")
    # ---------

    def test_AdditionBlock(self):
        pydtnn_layers = [
            AdditionBlock(
                [Conv2D(grouping=Conv2D.Grouping.STANDARD, nfilters=CONV2D_N_FILTERS, filter_shape=CONV2D_FILTER_SHAPE,
                        padding=CONV2D_PADDING, stride=CONV2D_STRIDE, dilation=CONV2D_DILATION),
                 BatchNormalization(gamma=BATCH_NORMALIZATION_GAMMA, beta=BATCH_NORMALIZATION_BETA, epsilon=BATCH_NORMALIZATION_EPSILON, momentum=BATCH_NORMALIZATION_MOMENTUM_PYDTNN)
                 ],

                [Conv2D(grouping=Conv2D.Grouping.STANDARD, nfilters=CONV2D_N_FILTERS, filter_shape=CONV2D_FILTER_SHAPE,
                        padding=CONV2D_PADDING, stride=CONV2D_STRIDE, dilation=CONV2D_DILATION)
                 ]
            )]

        torch_model = TorchAdditionBlock()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="AdditionBlock", rtol=2, atol=2)
    # ---------

    def test_ConcatenationBlock(self):
        pydtnn_layers = [
            ConcatenationBlock(
                [Conv2D(grouping=Conv2D.Grouping.STANDARD, nfilters=CONV2D_N_FILTERS, filter_shape=CONV2D_FILTER_SHAPE,
                        padding=CONV2D_PADDING, stride=CONV2D_STRIDE, dilation=CONV2D_DILATION),
                 BatchNormalization(gamma=BATCH_NORMALIZATION_GAMMA, beta=BATCH_NORMALIZATION_BETA, epsilon=BATCH_NORMALIZATION_EPSILON, momentum=BATCH_NORMALIZATION_MOMENTUM_PYDTNN)
                 ],

                [Conv2D(grouping=Conv2D.Grouping.STANDARD, nfilters=CONV2D_N_FILTERS, filter_shape=CONV2D_FILTER_SHAPE,
                        padding=CONV2D_PADDING, stride=CONV2D_STRIDE, dilation=CONV2D_DILATION)
                 ]
            )]
        torch_model = TorchConcatenationBlock()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="ConcatenationBlock", rtol=2, atol=2)
    # ---------

    def test_Sigmoid(self):
        pydtnn_layers = [Sigmoid()]
        torch_model = torch.nn.Sigmoid()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Sigmoid")
    # ---------

    def test_Relu(self):
        pydtnn_layers = [Relu()]
        torch_model = torch.nn.ReLU()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Relu")
    # ---------

    def test_Relu6(self):
        pydtnn_layers = [Relu6()]
        torch_model = torch.nn.ReLU6()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Relu6")
    # ---------

    def test_LeakyRelu(self):
        pydtnn_layers = [LeakyRelu()]
        torch_model = torch.nn.LeakyReLU()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="LeakyRelu")
    # ---------

    def test_Tanh(self):
        pydtnn_layers = [Tanh()]
        torch_model = torch.nn.Tanh()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Tanh")
    # ---------

    def test_Arctanh(self):
        # NOTE: Domain ArcTanH: Real numbers between "]-1, 1["
        pydtnn_layers = [Arctanh()]
        torch_model = TorchArcTanH()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Arctanh")
    # ---------

    def test_Log(self):
        pydtnn_layers = [Log()]
        torch_model = torch.nn.LogSigmoid()
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        # _x = np.where(_x < 0, 1, _x)
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Log")
    # ---------

    def test_Softmax(self):
        pydtnn_layers = [Softmax()]
        torch_model = torch.nn.Softmax(dim=1)
        pydtnn_model = LayerPyTorchTestCase.initialize_pydtnn_model(pydtnn_layers, params=self.params)
        _x = LayerPyTorchTestCase.get_test_data()
        self.do_test(_x=_x, pydtnn_model=pydtnn_model, torch_model=torch_model, name_test="Softmax")
    # ---------
    # ====================



import warnings
import numpy as np

from pydtnn.backends.cpu.utils.sparse_cython import \
    top_threshold_selection_dense_cython, \
    top_threshold_selection_coo_cython, \
    summ_coo_cython


class SparseMatrixCOO:
    """
    Represents a sparse matrix in COO format.
    This format stores the matrix using three arrays:
        - data: the nonzero values.
        - row: the row indices corresponding to each value.
        - col: the column indices corresponding to each value.
    The matrix is assumed to be in canonical format: indices sorted by row and then by column, and no duplicate entries are present.
    This class is not designed to store explict zeros so, len(self.data) should always be equal to nnz. 
    """

    def __init__(self, data, row, col, shape, has_canonical_format):
        """
        Primary initializer for SparseMatrixCOO.

        Parameters:
            data (np.ndarray): Array with the nonzero values.
            row  (np.ndarray): Array with the row indices.
            col  (np.ndarray): Array with the column indices.
            shape (tuple): Shape of the original matrix.
        """

        if len(data) != len(row) or len(data) != len(col):
            raise AssertionError("Data, row, and col arrays must have the same shape")

        if has_canonical_format:
            self.data = data
            self.row = row
            self.col = col
            self.shape = shape
            self.nnz = len(self.data)
            self.has_canonical_format = True
            assert (self._has_canonical_format())

        else:
            # TODO: order arrays in canonical format
            raise NotImplementedError("Not yet implemented constructor with unordered rows and cols")

    @classmethod
    def from_dense(cls, dense_array):
        """
        Alternative constructor to create a SparseMatrixCOO from a dense array.
        Only stores non-zero values!

        Parameters:
            dense_array (np.ndarray): A 2D dense matrix.

        Returns:
            SparseMatrixCOO: The sparse matrix in COO format
        """

        if len(dense_array.shape) != 2:
            raise AssertionError("Dense array must be 2D.")

        warnings.warn("From dense constructor should be used only in case of debugging for performance reasons.")

        row, col = np.where(dense_array != 0)
        data = dense_array[row, col]
        row = row.astype(np.int32)
        col = col.astype(np.int32)
        return cls(data, row, col, dense_array.shape, has_canonical_format=True)

    @classmethod
    def from_dense_top_selection(cls, dense_array, threshold):
        """
        Alternative constructor to create a SparseMatrixCOO from a dense array,
        considering only elements with absolute value greater than or equal to the threshold.

        Parameters:
            dense_array (np.ndarray): A 2D dense matrix.
            threshold (float): Threshold for including an element.

        Returns:
            SparseMatrixCOO: The sparse matrix in COO format, containing only significant elements.
        """

        if len(dense_array.shape) != 2:
            raise AssertionError("Dense array must be 2D.")

        # topk_row, topk_col = np.where(np.abs(dense_array) >= threshold)
        # topk = dense_array[topk_row, topk_col]
        topk, topk_row, topk_col = top_threshold_selection_dense_cython(dense_array, threshold)
        return cls(topk, topk_row, topk_col, dense_array.shape, has_canonical_format=True)

    def top_selection(self, threshold, inplace=True):
        """
        Performs top threshold selection on sparse array

        Parameters:
            threshold (float): Threshold for including an element.
            inplace (bool, optional):

        Returns:
            topk (SparseMatrixCOO): if inplace == False, or void (None): if inplace == True 
        """

        topk, topk_row, topk_col = top_threshold_selection_coo_cython(self.data, self.row, self.col, threshold)

        if inplace:
            self.data = topk
            self.row = topk_row
            self.col = topk_col
            self.nnz = len(self.data)
            # self.shape remains equal
            # self.has_canonical_format remains equal
        else:
            return SparseMatrixCOO(topk, topk_row, topk_col, self.shape, self.has_canonical_format)

    def get_indexes(self):
        """
        Returns the row and col

        Returns:
            tuple (tuple): row and col
        """
        return self.row, self.col

    def get_triplet(self):
        """
        Returns the data, row, col triplet

        Returns:
            triplet (tuple): data, row col
        """
        return self.data, self.row, self.col

    def slice(self, row_start, row_end, reset_indexes=False):
        """
        Perform a slice by row of the sparse matrix.

        Parameters:
            row_start (int): The starting row index (inclusive) of the slice.
            row_end (int): The ending row index (exclusive) of the slice.
            reset_indexes (bool, optional): If True, resets the row indices of the 
                                            sliced matrix so that `row_start` maps to zero.
                                            Defaults to False.
        Returns:
            coo_sliced: (SparseMatrixCOO): A row-sliced sparse matrix of self
        """
        start_index = np.searchsorted(self.row, row_start, side='left')
        ending_index = np.searchsorted(self.row, row_end, side='left')

        sliced_data = self.data[start_index:ending_index]
        sliced_row = self.row[start_index:ending_index]
        sliced_col = self.col[start_index:ending_index]
        if reset_indexes:
            sliced_row -= row_start

        return SparseMatrixCOO(sliced_data, sliced_row, sliced_col, self.shape, self.has_canonical_format)

    def to_dense(self):
        """
        Convert to dense np.array

        Returns:
            dense_matrix: (np.array): a dense matrix
        """

        warnings.warn("This function ('to_sparse') should be used only in case of debugging for performance reasons.")

        dense_matrix = np.zeros(self.shape, dtype=np.float32)
        dense_matrix[self.row, self.col] = self.data
        return dense_matrix

    def __add__(self, other):
        """
        Adds two SparseMatrixCOO matrices that are in canonical format.

        Parameters:
            other (SparseMatrixCOO): Another SparseMatrixCOO instance.

        Returns:
            SparseMatrixCOO: A new instance representing the sum of both matrices.
        """

        if type(other) != SparseMatrixCOO:
            raise AssertionError("Operand must be a SparseMatrixCOO instance.")
        if self.shape != other.shape:
            raise AssertionError("Matrices must have the same shape.")
        if not self.has_canonical_format or not other.has_canonical_format:
            raise AssertionError("Both matrices must be in canonical format.")

        summ_val, summ_row, summ_col = summ_coo_cython(self.data, self.row, self.col, other.data, other.row, other.col)
        return SparseMatrixCOO(summ_val, summ_row, summ_col, self.shape, has_canonical_format=True)

    def __radd__(self, other):
        """
        Implements right-hand addition to support the built-in sum() function.

        This method allows an instance of this class to be used with sum() by handling the
        case where the left operand is 0. If 'other' is 0, it returns the instance itself;
        otherwise, it delegates the operation to the __add__ method.

        Parameters:
            other (int or instance of the same class): The left-hand operand, typically 0 when used with sum().

        Returns:
            An instance of the class representing the sum of self and other.
        """
        if other == 0:
            return self
        return self.__add__(other)

    def _has_canonical_format(self):
        """
        Check if SparseMatrixCOO follows canonical format: 
            - Indexes are sorted by row and then by column 
            - There are no duplicate entries
            - There may have explicit zero elements
        This function is computationally expensive and therefore should only be used for developing/debugging purposes.
        This function should only be used in developement to assert that sparse matrices have canonical format. 

        Returns:
            has_canonical_format (boolean): True if indexes are in canonical format, False if not. 
        """

        warnings.warn("This function ('has_canonical_format') should be used only in case of debugging for performance reasons.")

        if self.nnz == 0:
            return True

        if not np.all(self.row[:-1] <= self.row[1:]):
            return False

        for i in range(self.nnz - 1):
            if self.row[i] == self.row[i + 1] and self.col[i] >= self.col[i + 1]:
                return False
        return True



import unittest
import numpy as np

import pycuda.gpuarray as gpuarray # type: ignore

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.layer_base import LayerBase
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.layer import LayerError
from pydtnn.model import Model
from pydtnn.tests.abstract.model_common import ModelCommonTestCase
from pydtnn.tests.abstract.common import verbose_test, Params
from pydtnn.utils.tensor import TensorFormat, format_transpose


class ModelGpuTestCase(ModelCommonTestCase):
    """
    Tests that two models with different parameters lead to the same results
    """
    global ModelCommonTestCase

    rtol_dict = ModelCommonTestCase.rtol_dict | {ConcatenationBlock: 1e-1, AdditionBlock: 1e-1, Conv2D: 1e-4}
    atol_dict = ModelCommonTestCase.atol_dict | {ConcatenationBlock: 1e-1, AdditionBlock: 1e-1, Conv2D: 1e-4}

    # NOTE: Delete parent test to prevent re-export and re-testing
    del ModelCommonTestCase

    # Compares results between an XX model {self.model1_desc} and other {self.model1_desc}
    model1_desc = "using the CPU backend"
    model2_desc = "using the GPU backend"

    def get_model2(self, model_name: str) -> Model:
        # GPU model
        params = Params()
        params.model_name = model_name  # type: ignore
        params.enable_gpu = True  # type: ignore
        params.enable_cudnn_auto_conv_alg = True  # type: ignore
        params.tensor_format = TensorFormat.NHWC.upper()
        params_dict = vars(params)
        try:
            model2 = Model(**params_dict)
        except LayerError as exc:
            raise unittest.SkipTest(f"Model {model_name} incompatible with {params_dict['dataset_name']}") from exc
        return model2

    def copy_weights_and_biases(self, model1: Model, model2: Model):
        """
        Copy weights and biases from Model 1 to Model 2
        """
        for cpu_layer, gpu_layer in zip(model1.get_all_layers(), model2.get_all_layers()):
            if cpu_layer.weights is None:
                continue
            if isinstance(gpu_layer, Conv2D):
                if model2.tensor_format is TensorFormat.NHWC:
                    gpu_layer.weights_cpu = format_transpose(cpu_layer.weights, "IHWO", "OHWI").copy()
                else:
                    gpu_layer.weights_cpu = cpu_layer.weights.copy()
            else:
                gpu_layer.weights_cpu = cpu_layer.weights.copy()
            if gpu_layer.weights_cpu is not None:
                weights_gpu = gpuarray.to_gpu(gpu_layer.weights_cpu)
                gpu_layer.weights = TensorGPU(weights_gpu, gpu_layer.model.tensor_format,
                                              gpu_layer.model.cudnn_dtype, TensorGPU.TensorTypeEnum.FILTER)
            if gpu_layer.use_bias:
                if cpu_layer.biases is None:
                    continue

                gpu_layer.biases_cpu = cpu_layer.biases.copy()
                if gpu_layer.biases_cpu is not None:
                    biases_gpu = gpuarray.to_gpu(gpu_layer.biases_cpu)
                    gpu_layer.biases = TensorGPU(biases_gpu, gpu_layer.model.tensor_format,
                                                 gpu_layer.model.cudnn_dtype)

    def set_data_to_ary(self, ary: "gpuarray",  # type: ignore
                        data: np.ndarray, layer: LayerBase) -> None:
        try:
            ary.set(data.copy())
        except ValueError as e:
            raise ValueError(f"Output of model 1 {layer.name_with_id}" \
                             f" is not ordered [x.strides: {data.strides}") from e
    # ----

    def do_model2_forward_pass(self, model2: Model, x1: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 2 forward pass
        """
        x2 = [x1[0]]
        # Input layer
        layer = model2.layers[0]
        self.set_data_to_ary(layer.y.ary, x1[0], layer)
        out = layer.forward(layer.y)
        x2.append(out.ary.get())

        # The rest of the layers
        for i, layer in enumerate(model2.layers[1:], 1):
            if verbose_test():
                print(layer)
            self.set_data_to_ary(model2.layers[i - 1].y.ary, x1[i], layer)
            out = layer.forward(model2.layers[i - 1].y)
            x2.append(out.ary.get())
        return x2

    def do_model2_backward_pass(self, model2: Model, dx1: list[np.ndarray]) -> list[np.ndarray]:
        """
        Model 2 backward pass
        """
        dx2 = [dx1[-1].copy()]

        layer = model2.layers[-1]
        self.set_data_to_ary(model2.layers[-1].dx.ary, dx1[-1], layer)
        out = layer.backward(model2.layers[-1].dx)
        dx2.insert(0, out.ary.get().copy())

        for i, layer in reversed(list(enumerate(model2.layers))[:-1]):
            if verbose_test():
                print(layer)
            self.set_data_to_ary(model2.layers[i + 1].dx.ary, dx1[i + 1], layer)
            out = layer.backward(model2.layers[i + 1].dx)
            dx2.insert(0, out.ary.get().copy())
        return dx2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Given two profiles extracted with extract_profile_from_output, prints the
differences between them.
"""

###########################################################################
# IMPORTS                                                                 #
###########################################################################
import getopt
import gzip
import pathlib
import sys
from prettytable import PrettyTable, MSWORD_FRIENDLY, PLAIN_COLUMNS


###########################################################################
# MISCELLANEOUS FUNCTIONS                                                 #
###########################################################################


def my_help():
    """Print the the command line usage help."""
    print("""Usage: diff_profiles.py [OPTION]... FILE1 FILE2

Given two profiles extracted with extract_profile_from_output, prints the
differences between them.

Options:
    -v, --VERBOSE      increment the output verbosity.
    -h, --help         display this help and exit.

Please, report bugs to <barrachi@uji.es>.
""")


def log(text):
    """Log a message to stderr."""
    sys.stderr.write(">>> %s\n" % text)


def error(text):
    """Report an error message and exit."""
    sys.stderr.write("ERROR: %s\n" % text)
    raise SystemExit(-1)


# Global command line parameters
INPUT_FILE_NAME = []
VERBOSE = 0
SCRIPT_PATH = pathlib.Path(__file__).parent.absolute()


def get_opts():
    """Read command line options."""
    global INPUT_FILE_NAME, VERBOSE
    optlist, args = getopt.getopt(sys.argv[1:],
                                  'hv',
                                  ['VERBOSE', 'help'])
    for opt, arg in optlist:
        if opt in ('-h', '--help'):
            my_help()
            raise SystemExit()
        elif opt in ('-v', '--verbosity'):
            VERBOSE = 1
    # Check required arguments
    if len(args) < 2:
        my_help()
        error("Two FILES are required")
    INPUT_FILE_NAME.append(args[0])
    INPUT_FILE_NAME.append(args[1])


###########################################################################
# APPLICATION SPECIFIC FUNCTIONS                                          #
###########################################################################
def file_to_dict(file):
    """Convert the CSV file to a dict."""
    _dict = {}
    total_time = 0.0
    # ncalls, tottime, percall, cumtime, percall, filename:lineno(function)
    file.readline()  # Ignore header
    for line in file.readlines():
        values = line.split(',', 5)
        for i in range(5):
            try:
                values[i] = int(values[i])
            except ValueError:
                try:
                    values[i] = float(values[i])
                except ValueError:
                    pass
        # Remove part of the path from values[5]
        splat_by_slash = values[5].split('/')
        values[5] = '/'.join(splat_by_slash[-3:])
        _dict[values[5]] = values[0:5]
        total_time += values[1]
    return _dict, total_time


def do_diff():
    """Do the diff"""
    data = []
    total_times = []
    print(f"Comparing '{INPUT_FILE_NAME[0]}' with '{INPUT_FILE_NAME[1]}'...")
    print()
    for i in range(2):
        file_open = open if INPUT_FILE_NAME[i][-3:] != ".gz" else gzip.open
        with file_open(INPUT_FILE_NAME[i], 'rt') as file:
            _dict, _total_time = file_to_dict(file)
            data.append(_dict)
            total_times.append(_total_time)
    t = PrettyTable(['ncalls', 'tottime', 'percall', 'cumtime', 'percall2', 'filename:lineno(function)'])
    t.set_style(PLAIN_COLUMNS)
    t.align = "r"
    t.align['filename:lineno(function)'] = "l"
    t.sortby = "tottime"
    t.reversesort = True
    common_keys = []
    for key, values1 in data[0].items():
        if key in data[1]:
            common_keys.append(key)
            values2 = data[1][key]
            try:
                values = [round(x - y, 3) for x, y in zip(values1, values2)]
            except TypeError:
                values = []
                for x, y in zip(values1, values2):
                    if isinstance(x, type(y)) == str:
                        x0, x1 = x.split('/')
                        y0, y1 = y.split('/')
                        # values.append(f"{int(x0)-int(y0)}/{int(x1)-int(y1)}")
                        values.append(int(x0) - int(y0))  # PrettyTable sort does not work with the previous version
                    else:
                        values.append(round(x - y, 3))
            if values == [0, 0.0, 0.0, 0.0, 0.0]:
                continue
            t.add_row([values[0], values[1], values[2], values[3], values[4], key])
    print(f"Differences between '{INPUT_FILE_NAME[0]}' and '{INPUT_FILE_NAME[1]}'")
    print(t)
    print()
    t.clear_rows()
    for common_key in common_keys:
        data[0].pop(common_key)
        data[1].pop(common_key)
    for i in range(2):
        if data[0]:
            print(f"Calls only in '{INPUT_FILE_NAME[i]}'")
            for key, values in data[i].items():
                t.add_row([values[0], values[1], values[2], values[3], values[4], key])
            print(t)
            print()
            t.clear_rows()
    print("Total times")
    t = PrettyTable([INPUT_FILE_NAME[0], INPUT_FILE_NAME[1]])
    t.align = "r"
    t.add_row([round(x, 3) for x in total_times])
    print(t)
    print()


def main():
    """Do the work (main function, called when not imported)."""
    get_opts()
    # Main part of the application
    do_diff()


if __name__ == "__main__":
    main()



from pydtnn.layers.abstract.block_layer import AbstractBlockLayer
from pydtnn.utils.constants import Array


class EncoderDecoder[T: Array](AbstractBlockLayer[T]):
    def __init__(self, enc_layers: int = 1, dec_layers: int = 1, embedl: int = 64, d_k: int = 3, heads: int = 10, d_ff: int = 256, dropout_rate: float = 0.5):
        super().__init__()
        self.embedl = embedl
        self.enc_layers = enc_layers
        self.dec_layers = dec_layers
        self.heads = heads
        self.d_k = d_k
        self.d_ff = d_ff
        self.dropout_rate = dropout_rate
        self.encoder = [None,]
        self.decoder = [None,]
        self.paths = [self.encoder + self.decoder]

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)

        if self.shape == ():
            self.shape = prev_shape[0]

    def _show_props(self) -> dict:
        props = super()._show_props()

        props["encodes"] = self.enc_layers
        props["decodes"] = self.dec_layers

        return props



"""
PyDTNN convWinograd module
"""

import ctypes
import math
import platform
from warnings import warn
import weakref
from collections import defaultdict
from functools import partial

import numpy as np
from pydtnn.backends.cpu.utils.im2col_nchw_cython import im2col_nchw_cython
from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython

from pydtnn.utils.tensor import TensorFormat, encode_shape, decode_shape
from pydtnn.utils import load_library
from pydtnn.utils.best_of import BestOf
from pydtnn.utils.memory_cache import MemoryCache

try:
    load_library("convwinograd")
    is_conv_winograd_available = True
except Exception:
    is_conv_winograd_available = False


class ConvWinograd:
    """
    Exposes the libconvWinograd functions following the PyDTNN conventions.

    Methods
    -------
    winograd(weights, x, biases, vpadding, hpadding,
             vstride, hstride, vdilation, hdilation)
        calls the appropriate winograd function from libconvWinograd.so to perform
        the Toom-Cook based convolution.

    Examples
    --------
    See __usage_example__() method for an example of use. This example can be
    run with: 'python conv_winograd.py'

    Tests
    -----
    To perform the tests, run the following command from the current directory:
        python -m unittest tests.convWinogradTestcase

    (see tests/winograd.py for more instructions on testing)
    """

    lib_cw = None  # will link to the libconvwinograd.so library

    def __init__(self, kh: int, kw: int, vstride: int, hstride: int, vdilation: int, hdilation: int,
                 dtype: np.dtype = np.dtype(np.float32), tensor_format=TensorFormat.NCHW,
                 debug=False, parent_layer=None):
        """
        Loads the libconvWinograd.so library.

        Parameters
        ----------
        kh : kernel height

        kw : kernel width

        vstride : vertical stride

        hstride : horizontal stride

        vdilation : vertical dilation

        hdilation : horizontal dilation

        dtype : data type
            The element data type being used on all the matrices.
        debug : boolean
            Whether to print debug information or not.
        parent_layer: layer
            The layer that is using it (for tracing purposes).
        """

        def register_winograd_function(m: int, r: int, g: np.ndarray, bt: np.ndarray, at: np.ndarray) -> None:
            # choose the appropriate convWinograd function depending on the architecture and the data type being used
            if platform.machine() == 'aarch64':
                if self.dtype == np.float32:
                    routine_names = [("neon", f"conv_winograd_{m}x{m}_{r}x{r}_neon_fp32_{self.tensor_format}")]
                else:
                    raise NotImplementedError(
                        f"Type {str(self.dtype)} not supported by this version of libconvWinograd!")
            elif platform.machine() == 'x86_64':
                if self.dtype == np.float32:
                    routine_names = [(intr, f"conv_winograd_{m}x{m}_{r}x{r}_{intr}_fp32_{self.tensor_format}")
                                     for intr in ["native", "sse", "avx", "avx512"]]
                else:
                    raise NotImplementedError(
                        f"Type {str(self.dtype)} not supported by this version of libconvWinograd!")
            else:
                raise NotImplementedError(f"Platform '{str(platform.machine())}' not yet supported")

            funcs = []
            for rn in routine_names:
                try:
                    funcs.append((
                        rn[0],
                        (
                            self._conv_winograd_c,
                            getattr(self.__class__.lib_cw, f"{rn[1]}_pre"),
                            getattr(self.__class__.lib_cw, f"{rn[1]}_kernel")
                        )
                    ))
                except AttributeError:
                    pass
            if not funcs:
                warn("Winograd routine not found. Fallback to numpy version!", RuntimeWarning)
                funcs = [("numpy", (self._conv_winograd_numpy, None, None))]

            for intr, f in funcs:
                self.alternatives[r].append((f"cw{m}{r}{intr}", partial(f[0], m, r, g, bt, at, f[1], f[2])))

        # Parent layer
        if parent_layer is not None:
            self.get_parent_layer = weakref.ref(parent_layer)
            self.evaluate_only = self.get_parent_layer().model.evaluate_only
            enable_best_of = self.get_parent_layer().model.enable_best_of
        else:
            enable_best_of = False
            self.evaluate_only = True

        if isinstance(dtype, np.dtype):
            self.dtype = dtype
        else:
            try:
                self.dtype = {'float32': np.float32, 'float64': np.float64}[dtype]
            except KeyError:
                raise NotImplementedError("dtype '{}' not recognized".format(dtype))

        self.tensor_format = tensor_format

        if ConvWinograd.lib_cw is None:
            ConvWinograd.lib_cw = load_library("convwinograd")

        self.alternatives = defaultdict(lambda: [])
        m, r = None, None

        if (kh, kw) == (2, 2) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
            # F(3x3, 2x2)
            m, r = 3, 2
            register_winograd_function(m, r,
                                       g=np.array([[1, 0],
                                                   [1. / 2., 1. / 2.],
                                                   [1. / 2., -1. / 2.],
                                                   [0, 1]],
                                                  dtype=self.dtype, order="C"),
                                       bt=np.array([[1, 0, -1, 0],
                                                    [0, 1, 1, 0],
                                                    [0, -1, 1, 0],
                                                    [0, -1, 0, 1]],
                                                   dtype=self.dtype, order="C"),
                                       at=np.array([[1, 1, 1, 0],
                                                    [0, 1, -1, 0],
                                                    [0, 1, 1, 1]],
                                                   dtype=self.dtype, order="C"))

        if (kh, kw) == (3, 3) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
            # F(2x2, 3x3)
            m, r = 2, 3
            register_winograd_function(m, r,
                                       g=np.array([[1, 0, 0],
                                                   [1. / 2., 1. / 2., 1. / 2.],
                                                   [1. / 2., -1. / 2., 1. / 2.],
                                                   [0, 0, 1]],
                                                  dtype=self.dtype, order="C"),
                                       bt=np.array([[1, 0, -1, 0],
                                                    [0, 1, 1, 0],
                                                    [0, -1, 1, 0],
                                                    [0, 1, 0, -1]],
                                                   dtype=self.dtype, order="C"),
                                       at=np.array([[1, 1, 1, 0],
                                                    [0, 1, -1, -1]],
                                                   dtype=self.dtype, order="C"))

        if (kh, kw) == (3, 3) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
            # F(4x4, 3x3)
            m, r = 4, 3
            register_winograd_function(m, r,
                                       g=np.array([[1. / 4., 0, 0],
                                                   [-1. / 6., -1. / 6., -1. / 6.],
                                                   [-1. / 6., 1. / 6., -1. / 6.],
                                                   [1. / 24., 1. / 12., 1. / 6.],
                                                   [1. / 24., -1. / 12., 1. / 6.],
                                                   [0, 0, 1]],
                                                  dtype=self.dtype, order="C"),
                                       bt=np.array([[4, 0, -5, 0, 1, 0],
                                                    [0, -4, -4, 1, 1, 0],
                                                    [0, 4, -4, -1, 1, 0],
                                                    [0, -2, -1, 2, 1, 0],
                                                    [0, 2, -1, -2, 1, 0],
                                                    [0, 4, 0, -5, 0, 1]],
                                                   dtype=self.dtype, order="C"),
                                       at=np.array([[1, 1, 1, 1, 1, 0],
                                                    [0, 1, -1, 2, -2, 0],
                                                    [0, 1, 1, 4, 4, 0],
                                                    [0, 1, -1, 8, -8, 1]],
                                                   dtype=self.dtype, order="C"))

        if (kh, kw) == (5, 5) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
            # F(2x2, 5x5)
            m, r = 2, 5
            register_winograd_function(m, r,
                                       g=np.array([[1. / 4., 0, 0, 0, 0],
                                                   [-1. / 6., -1. / 6., -1. / 6., -1. / 6., -1. / 6.],
                                                   [-1. / 6., 1. / 6., -1. / 6., 1. / 6., -1. / 6.],
                                                   [1. / 24., 1. / 12., 1. / 6., 1. / 3., 2. / 3.],
                                                   [1. / 24., -1. / 12., 1. / 6., -1. / 3., 2. / 3.],
                                                   [0, 0, 0, 0, 1]],
                                                  dtype=self.dtype, order="C"),
                                       bt=np.array([[4, 0, -5, 0, 1, 0],
                                                    [0, -4, -4, 1, 1, 0],
                                                    [0, 4, -4, -1, 1, 0],
                                                    [0, -2, -1, 2, 1, 0],
                                                    [0, 2, -1, -2, 1, 0],
                                                    [0, 4, 0, -5, 0, 1]],
                                                   dtype=self.dtype, order="C"),
                                       at=np.array([[1, 1, 1, 1, 1, 0],
                                                    [0, 1, -1, 2, -2, 1]],
                                                   dtype=self.dtype, order="C"))

        if r not in self.alternatives:
            raise NotImplementedError(f"Winograd not implemented for kernel {kh}x{kw}")

        try:
            self.conv_winograd_workspace_alloc_pre = \
                getattr(self.__class__.lib_cw, f"conv_winograd_workspace_alloc_pre")
            self.conv_winograd_workspace_alloc_kernel = \
                getattr(self.__class__.lib_cw, f"conv_winograd_workspace_alloc_kernel")
        except AttributeError:
            print(f"Winograd conv_winograd_workspace_alloc_pre/kernel routines not found.")
            self.conv_winograd_workspace_alloc_pre = None
            self.conv_winograd_workspace_alloc_kernel = None

        def winograd_workspace_alloc_pre(m, r, k, c):
            _u = ctypes.POINTER(ctypes.c_float)()
            self.conv_winograd_workspace_alloc_pre(ctypes.c_uint(m), ctypes.c_uint(r),
                                                   ctypes.c_uint(k), ctypes.c_uint(c), ctypes.byref(_u))
            return np.array([False]), _u

        def winograd_workspace_alloc_kernel(m, r, n, k, c, hi, wi, kh, kw, vpadding, hpadding):
            _v = ctypes.POINTER(ctypes.c_float)()
            _m = ctypes.POINTER(ctypes.c_float)()
            self.conv_winograd_workspace_alloc_kernel(ctypes.c_uint(m), ctypes.c_uint(r),
                                                      ctypes.c_uint(n), ctypes.c_uint(k), ctypes.c_uint(c),
                                                      ctypes.c_uint(hi), ctypes.c_uint(wi), ctypes.c_uint(kh), ctypes.c_uint(kw),
                                                      ctypes.c_uint(vpadding), ctypes.c_uint(hpadding), ctypes.byref(_v), ctypes.byref(_m))
            return _v, _m

        self.cw_cache_pre = MemoryCache(lambda args: winograd_workspace_alloc_pre(*args))
        self.cw_cache_kernel = MemoryCache(lambda args: winograd_workspace_alloc_kernel(*args))
        self.y_cache = MemoryCache(lambda shape: np.zeros(shape, self.dtype, order="C"))
        self.d_cache = MemoryCache(lambda shape: np.zeros(shape, self.dtype, order="C"))

        # Debug
        self.debug = debug

        self._reuse_processed_weights = False
        if self.evaluate_only:
            self._reuse_processed_weights = True
        self._weights_already_processed = False

        if enable_best_of and len(self.alternatives[r]) > 1:
            setattr(self, f"conv_winograd_{self.tensor_format}", BestOf(
                name="Winograd functions",
                alternatives=self.alternatives[r],
                get_problem_size=lambda *args, **kwargs: tuple(list(args[0].shape) + list(args[1].shape[1:])),
            ))
        else:
            setattr(self, f"conv_winograd_{self.tensor_format}", self.alternatives[r][0][1])

    def encode_shape(self, shape):
        return encode_shape(shape, self.tensor_format)

    def decode_shape(self, shape):
        return decode_shape(shape, self.tensor_format)

    def _conv_winograd_numpy(self, m: int, r: int, g: np.ndarray, bt: np.ndarray, at: np.ndarray, pre, kernel,
                             weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None, vpadding=0, hpadding=0,
                             vstride=1, hstride=1, vdilation=1, hdilation=1,
                             relu=False, bn=False,
                             running_mean: np.ndarray | None = None, inv_std: np.ndarray | None = None,
                             gamma: np.ndarray | None = None, beta: np.ndarray | None = None) -> np.ndarray:

        n, ci, hi, wi = self.decode_shape(x.shape)

        if self.tensor_format == TensorFormat.NCHW:
            co, ci, kh, kw = weights.shape
        else:
            ci, kh, kw, co = weights.shape

        t = m + r - 1  # Winograd sliding window size t x t
        s = m  # Winograd sliding window stride: t - (r - 1) = m

        if (kh, kw) != (r, r):
            raise ValueError(f"Kernel size {kh}x{kw} not supported by this version of Winograd, "
                             f"kernel size should be ({r}x{r})!")

        if (vstride, hstride) != (1, 1):
            raise ValueError(f"Stride {vstride}x{hstride} supported by this version of Winograd, "
                             f"stride should be (1,1)!")

        if (vdilation, hdilation) != (1, 1):
            raise ValueError(f"Dilation {vdilation}x{hdilation} supported by this version of Winograd, "
                             f"dilation should be (1,1)!")

        ho = (hi + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
        wo = (wi + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

        tile_h = math.ceil((hi + 2 * vpadding - t) / s) + 1
        tile_w = math.ceil((wi + 2 * hpadding - t) / s) + 1

        y_shape = self.encode_shape((n, co, ho, wo))
        y = self.y_cache[y_shape]
        u = np.zeros((t, t, co, ci), self.dtype, order="C")                     # FIXME: self.u_cache[(t, t, co, ci)]  # Workspace for G * g * G^T
        v = np.zeros((t, t, ci, (n * tile_h * tile_w)), self.dtype, order="C")  # FIXME: self.v_cache[(t, t, ci, (n * tile_h * tile_w))]
        # m_= self.m_cache[(t, t, co, (n * tile_h * tile_w))]
        d = self.d_cache[(t, t)]

        for k in range(co):
            for c in range(ci):
                # U = G  * g * G^T
                u[..., k, c] = (g @ weights[k, c, ...]) @ g.T

        # 1.1) First alternative: padding first
        # x_padded = best_pad(x, vpadding, hpadding)
        # _, _, hi, wi = x_padded.shape
        #
        # for c in range(ci):
        #     for b in range(n):
        #         for h in range(tile_h):
        #             for w in range(tile_w):
        #                 hh, ww = h * s, w * s
        #                 th, tw = min(hi-hh,t), min(wi-ww,t)
        #                 d[:th, :tw] = x_padded[b, c, hh:hh+th, ww:ww+tw]
        #                 v[..., c, b * tile_h * tile_w + h * tile_w + w] = (self.bt @ d) @ self.bt.T

        # 1.2) Second alternative: avoid padding
        for c in range(ci):
            for b in range(n):
                for h in range(tile_h):
                    hh_ = min(hi, h * s - vpadding)
                    hh, fh = (hh_, 0) if hh_ > 0 else (0, min(-hh_, t))
                    oh = max(min(t, hi - hh) - min(t, fh), 0)

                    for w in range(tile_w):
                        ww_ = min(wi, w * s - hpadding)
                        ww, fw = (ww_, 0) if ww_ > 0 else (0, min(-ww_, t))
                        ow = max(min(t, wi - ww) - min(t, fw), 0)

                        if hh < hh + oh and ww < ww + ow:
                            match self.tensor_format:
                                case TensorFormat.NCHW:
                                    d[fh:fh + oh, fw:fw + ow] = x[b, c, hh:hh + oh, ww:ww + ow]
                                case TensorFormat.NHWC:
                                    d[fh:fh + oh, fw:fw + ow] = x[b, hh:hh + oh, ww:ww + ow, c]
                                case tensor_format:
                                    raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")

                        #   0  0  0
                        #   X  X  X
                        #   X  X  X
                        if 0 <= fh:
                            d[:fh, ...] = 0

                        #   0  0  0
                        #   X  X  X
                        #   0  0  0
                        if fh + oh < t:
                            d[fh + oh:, ...] = 0

                        #   0  0  0
                        #   0  X  X
                        #   0  0  0
                        if 0 <= fw:
                            d[fh:fh + oh, :fw] = 0

                        #   0  0  0
                        #   0  X  0
                        #   0  0  0
                        if fw + ow < t:
                            d[fh:fh + oh, fw + ow:] = 0

                        v[..., c, b * tile_h * tile_w + h * tile_w + w] = (bt @ d) @ bt.T

        # 2.1) First alternative: np.einsum
        m_ = np.einsum('... i j, ... j k -> ... i k', u, v)

        # 2.2) Second alternative: matmul
        # for i in range(t):
        #     for j in range(t):
        #         m_[i, j] = u[i, j] @ v[i, j]

        for k in range(co):
            for b in range(n):
                for h in range(tile_h):
                    for w in range(tile_w):
                        z = (at @ m_[..., k, b * tile_h * tile_w + h * tile_w + w]) @ at.T
                        hh, ww = h * s, w * s
                        match self.tensor_format:
                            case TensorFormat.NCHW:
                                y[b, k, hh:hh + m, ww:ww + m] = z[:min(m, ho - hh), :min(m, wo - ww)]
                            case TensorFormat.NHWC:
                                y[b, hh:hh + m, ww:ww + m, k] = z[:min(m, ho - hh), :min(m, wo - ww)]
                            case tensor_format:
                                raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")

            if biases is not None:
                match self.tensor_format:
                    case TensorFormat.NCHW:
                        y[:, k, ...] += biases[k]
                    case TensorFormat.NHWC:
                        y[..., k] += biases[k]
                    case tensor_format:
                        raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")

            if bn:
                match self.tensor_format:
                    case TensorFormat.NCHW:
                        y[:, k, ...] = (((y[:, k, ...] - running_mean[k]) * inv_std[k]) * gamma[k]) + beta[k]
                    case TensorFormat.NHWC:
                        y[..., k] = (((y[..., k] - running_mean[k]) * inv_std[k]) * gamma[k]) + beta[k]
                    case tensor_format:
                        raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")

        if relu:
            y[y < 0] = 0

        return y

    def _conv_winograd_c(self, m: int, r: int, g: np.ndarray, bt: np.ndarray, at: np.ndarray,
                         x_winograd_pre, x_winograd_kernel,
                         weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None,
                         vpadding=0, hpadding=0, vstride=1, hstride=1, vdilation=1, hdilation=1,
                         relu=False, bn=False,
                         running_mean: np.ndarray | None = None, inv_std: np.ndarray | None = None,
                         gamma: np.ndarray | None = None, beta: np.ndarray | None = None) -> np.ndarray:

        n, ci, hi, wi = self.decode_shape(x.shape)

        match self.tensor_format:
            case TensorFormat.NCHW:
                co, ci, kh, kw = weights.shape
            case TensorFormat.NHWC:
                ci, kh, kw, co = weights.shape
            case tensor_format:
                raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")

        t = m + r - 1  # Winograd sliding window size t x t
        s = m  # Winograd sliding window stride: t - (r - 1) = m

        if (kh, kw) != (r, r):
            raise ValueError(f"Kernel size {kh}x{kw} not supported by this version of Winograd, "
                             f"kernel size should be ({r}x{r})!")

        if (vstride, hstride) != (1, 1):
            raise ValueError(f"Stride {vstride}x{hstride} supported by this version of Winograd, "
                             f"stride should be (1,1)!")

        if (vdilation, hdilation) != (1, 1):
            raise ValueError(f"Dilation {vdilation}x{hdilation} supported by this version of Winograd, "
                             f"dilation should be (1,1)!")

        ho = (hi + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
        wo = (wi + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

        _weights_already_processed, _u, = self.cw_cache_pre[(m, r, co, ci)]
        _v, _m = self.cw_cache_kernel[(m, r, n, co, ci, hi, wi, kh, kw, vpadding, hpadding)]

        y_shape = self.encode_shape((n, co, ho, wo))
        y = self.y_cache[y_shape]

        match self.tensor_format:
            case TensorFormat.NCHW:
                ld_d1, ld_d2, ld_d3 = ci * hi * wi, hi * wi, wi
                ld_f1, ld_f2, ld_f3 = ci * kh * kw, kh * kw, kw
                ld_y1, ld_y2, ld_y3 = co * ho * wo, ho * wo, wo
            case TensorFormat.NHWC:
                ld_d1, ld_d2, ld_d3 = hi * wi * ci, wi * ci, ci
                ld_f1, ld_f2, ld_f3 = kh * kw * co, kw * co, co
                ld_y1, ld_y2, ld_y3 = ho * wo * co, wo * co, co
            case tensor_format:
                raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")

        if not self._reuse_processed_weights or not _weights_already_processed[0]:
            x_winograd_pre(ctypes.c_uint(m), ctypes.c_uint(r),
                           ctypes.c_uint(n), ctypes.c_uint(co), ctypes.c_uint(ci),
                           ctypes.c_uint(kh), ctypes.c_uint(kw),
                           ctypes.c_void_p(weights.ctypes.data),
                           ctypes.c_uint(ld_f1), ctypes.c_uint(ld_f2), ctypes.c_uint(ld_f3), _u)
            _weights_already_processed[0] = True

        x_winograd_kernel(ctypes.c_uint(m), ctypes.c_uint(r),
                          ctypes.c_uint(n), ctypes.c_uint(co), ctypes.c_uint(ci),
                          ctypes.c_uint(hi), ctypes.c_uint(wi),
                          ctypes.c_uint(kh), ctypes.c_uint(kw),
                          ctypes.c_uint(vpadding), ctypes.c_uint(hpadding),
                          ctypes.c_void_p(x.ctypes.data), ctypes.c_uint(ld_d1), ctypes.c_uint(ld_d2), ctypes.c_uint(ld_d3),
                          ctypes.c_void_p(y.ctypes.data), ctypes.c_uint(ld_y1), ctypes.c_uint(ld_y2), ctypes.c_uint(ld_y3),
                          ctypes.c_void_p(None if biases is None else biases.ctypes.data), _u, _v, _m,
                          ctypes.c_char((b'F', b'T')[relu]), ctypes.c_char((b'F', b'T')[bn]),
                          ctypes.c_void_p(None if running_mean is None else running_mean.ctypes.data),
                          ctypes.c_void_p(None if inv_std is None else inv_std.ctypes.data),
                          ctypes.c_void_p(None if gamma is None else gamma.ctypes.data),
                          ctypes.c_void_p(None if beta is None else beta.ctypes.data))
        return y


def time_it_func(x: np.ndarray, w_c: np.ndarray, biases: np.ndarray,
                 b: int, kn: int,
                 ho: int, wo: int, kh: int, kw: int,
                 vpadding: int, hpadding: int, vstride: int, hstride: int,
                 vdilation: int, hdilation: int,
                 ) -> int | float:

    res = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2row_nhwc_cython(x, res,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride,
                       vdilation, hdilation)
    res = res @ w_c
    res += biases.reshape(b * ho * wo, kn)
    return res


def time_it_im2col(x: np.ndarray, w_c: np.ndarray, biases: np.ndarray,
                   b: int, kn: int,
                   ho: int, wo: int, kh: int, kw: int,
                   vpadding: int, hpadding: int, vstride: int, hstride: int,
                   vdilation: int, hdilation: int,
                   ) -> int | float:

    res = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2col_nchw_cython(x, res,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride,
                       vdilation, hdilation)
    res = res @ w_c
    res += biases.reshape(b * ho * wo, kn)
    return res


def time_it_im2col_4_dims(x: np.ndarray, w_c: np.ndarray, biases: np.ndarray,
                          kk: int, ho: int, wo: int, kh: int, kw: int,
                          vpadding: int, hpadding: int, vstride: int, hstride: int,
                          vdilation: int, hdilation: int,
                          ) -> int | float:

    res = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2col_nchw_cython(x, res,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride,
                       vdilation, hdilation)
    res = res @ w_c
    res += biases.reshape(kk, -1, ho, wo).transpose(1, 0, 2, 3)
    return res


def __usage_example__():
    # Imports for this usage example (not required otherwise)
    from timeit import timeit
    from pydtnn.utils import random
    # Default parameters (1st layer AlexNet for Cifar10)
    b = 32  # Batch size
    c = 16  # Channels per layer
    h = 18  # Layers height
    w = 18  # Layers width
    kn = 16  # Number of filters
    kh = 3  # Filters weights height
    kw = 3  # Filters weights width
    vpadding = 1  # Vertical padding
    hpadding = 1  # Horizontal padding
    vstride = 1  # Vertical stride
    hstride = 1  # Horizontal stride
    vdilation = 1  # Vertical dilation
    hdilation = 1  # Horizontal dilation
    # Create weights, x, and biases matrices from previous parameters. If no biases
    # matrix is provided, a proper one filled with zeros will be automatically
    # created.
    random.seed(0)
    # weights[1][1][1][1] = -322.0
    # weights[2][2][2][2] = -334.0

    ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
    wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

    biases_wg = (np.ones(kn) * 10).astype(np.float32, order='C')

    # NCHW --------------------------
    weights = random.random((kn, c, kh, kw)).astype(np.float32, order='C')
    x = random.random((b, c, h, w)).astype(np.float32, order='C')
    biases = (np.ones((kn, b * ho * wo)) * 10).astype(np.float32, order='C')
    print("Using conv_winograd NCHW to compute weights * x + biases...")
    r = False
    conv_winograd = ConvWinograd(kh, kw, vstride, hstride, vdilation, hdilation, debug=False)
    conv_winograd_result_nchw = conv_winograd.conv_winograd_nchw(weights, x, biases_wg,
                                                                 vpadding=vpadding, hpadding=hpadding,
                                                                 vstride=vstride, hstride=hstride,
                                                                 vdilation=vdilation, hdilation=hdilation, relu=r)
    conv_winograd_t = timeit(lambda: conv_winograd.conv_winograd_nchw(weights, x, biases_wg,
                                                                      vpadding=vpadding, hpadding=hpadding,
                                                                      vstride=vstride, hstride=hstride,
                                                                      vdilation=vdilation, hdilation=hdilation, relu=r),
                             number=10) / 10
    print("conv_winograd time: {:.4f}".format(conv_winograd_t))
    print("Using im2col and mm NCHW ...")
    x_c = np.zeros((c * kh * kw, b * ho * wo))
    im2col_nchw_cython(x, x_c,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride, vdilation, hdilation)
    w_c = weights.reshape(kn, -1)
    im2col_mm_result_nchw = (w_c @ x_c + biases).reshape(kn, -1, ho, wo).transpose(1, 0, 2, 3)
    mm_t = timeit(lambda: time_it_im2col(x, w_c, biases, b, kn, ho, wo, kh, kw, vpadding, hpadding, vstride, hstride, vdilation, hdilation),
                  number=10) / 10
    print("mm time: {:.4f}".format(mm_t))

    # NHWC --------------------------
    weights = random.random((c, kh, kw, kn)).astype(np.float32, order='C')
    x = random.random((b, h, w, c)).astype(np.float32, order='C')
    biases = (np.ones((b * ho * wo, kn)) * 10).astype(np.float32, order='C')
    print("Using conv_winograd NHWC to compute weights * x + biases...")
    r = False
    conv_winograd = ConvWinograd(kh, kw, vstride, hstride, vdilation, hdilation,
                                 tensor_format=TensorFormat.NHWC, debug=False)
    conv_winograd_result_nhwc = conv_winograd.conv_winograd_nhwc(weights, x, biases_wg,
                                                                 vpadding=vpadding, hpadding=hpadding,
                                                                 vstride=vstride, hstride=hstride,
                                                                 vdilation=vdilation, hdilation=hdilation, relu=r)
    conv_winograd_t = timeit(lambda: conv_winograd.conv_winograd_nhwc(weights, x, biases_wg,
                                                                      vpadding=vpadding, hpadding=hpadding,
                                                                      vstride=vstride, hstride=hstride,
                                                                      vdilation=vdilation, hdilation=hdilation, relu=r),
                             number=10) / 10
    print("conv_winograd time: {:.4f}".format(conv_winograd_t))
    print("Using im2col and mm NHWC ...")
    x_c = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2row_nhwc_cython(x, x_c,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride, vdilation, hdilation)
    w_c = weights.reshape((-1, kn), copy=False)
    im2col_mm_result_nhwc = (x_c @ w_c + biases).reshape(-1, ho, wo, kn)
    mm_t = timeit(
        lambda: time_it_func(x, w_c, biases, b, kn, ho, wo, kh, kw, vpadding, hpadding, vstride, hstride, vdilation, hdilation),
        number=10) / 10
    print("mm time: {:.4f}".format(mm_t))

    if r:
        im2col_mm_result_nhwc = np.maximum(im2col_mm_result_nhwc, 0)

    print("Sum WINOGRAD NCHW: ", conv_winograd_result_nchw.sum(), conv_winograd_result_nchw.shape)
    print("Sum   IM2COL NCHW: ", im2col_mm_result_nchw.sum(), im2col_mm_result_nchw.shape)
    print("np.allclose NCHW: ", np.allclose(conv_winograd_result_nchw, im2col_mm_result_nchw, atol=1e-3))
    print()
    print("Sum WINOGRAD NHWC: ", conv_winograd_result_nhwc.sum(), conv_winograd_result_nhwc.shape)
    print("Sum   IM2COL NHWC: ", im2col_mm_result_nhwc.sum(), im2col_mm_result_nhwc.shape)
    print("np.allclose NHWC: ", np.allclose(conv_winograd_result_nhwc, im2col_mm_result_nhwc, atol=1e-3))
    # """

    # n = 65
    # c = k = 65
    # h = w = 33
    # vpadd = hpadd = 6
    n = 17
    c = k = 17
    h = w = 33
    vpadd = hpadd = 6
    for nn in range(16, n, 16):
        for cc in range(16, c, 16):
            for kk in range(16, k, 16):
                for hh in range(8, h, 4):
                    for vpadding in range(1, vpadd):
                        for hpadding in range(1, hpadd):
                            for kh in [2, 3, 5]:
                                kw = kh
                                ww = hh
                                ho = (hh + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
                                wo = (ww + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

                                for tensor_fmt in [TensorFormat.NCHW, TensorFormat.NHWC]:
                                    conv_winograd = ConvWinograd(kh, kw, vstride, hstride, vdilation, hdilation,
                                                                 tensor_format=tensor_fmt, debug=False)
                                    print(nn, cc, kk, hh, ww, vpadding, hpadding, kh, conv_winograd.tensor_format_str,
                                          end="")

                                    biases_wg = (np.ones(kk) * 10).astype(np.float32, order='C')
                                    match tensor_fmt:
                                        case TensorFormat.NCHW:
                                            weights = random.random((kk, cc, kh, kw)).astype(np.float32, order='C')
                                            x = random.random((nn, cc, hh, ww)).astype(np.float32, order='C')
                                            biases = (np.ones((kk, nn * ho * wo)) * 10).astype(np.float32, order='C')
                                            w_c = weights.reshape(kk, -1)
                                            res = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
                                            im2col_nchw_cython(x, res,
                                                               kh, kw, ho, wo,
                                                               vpadding, hpadding, vstride, hstride,
                                                               vdilation, hdilation)
                                            im2col_mm_result = (w_c @ res) + biases
                                            im2col_mm_result = \
                                                im2col_mm_result.reshape(kk, -1, ho, wo).transpose(1, 0, 2, 3)
                                            im2col_t = timeit(lambda: time_it_im2col_4_dims(x, w_c, biases, kk, ho, wo, kh, kw, vpadding, hpadding, vstride, hstride, vdilation, hdilation),
                                                              number=10) / 10

                                            conv_winograd_result = conv_winograd.conv_winograd_nchw(weights, x, biases_wg,
                                                                                                    vpadding=vpadding,
                                                                                                    hpadding=hpadding,
                                                                                                    vstride=vstride,
                                                                                                    hstride=hstride,
                                                                                                    vdilation=vdilation,
                                                                                                    hdilation=hdilation)
                                            conv_winograd_t = timeit(
                                                lambda: conv_winograd.conv_winograd_nchw(weights, x, biases_wg,
                                                                                         vpadding=vpadding,
                                                                                         hpadding=hpadding,
                                                                                         vstride=vstride, hstride=hstride,
                                                                                         vdilation=vdilation,
                                                                                         hdilation=hdilation),
                                                number=10) / 10
                                        case TensorFormat.NHWC:
                                            weights = random.random((cc, kh, kw, kk)).astype(np.float32, order='C')
                                            x = random.random((nn, hh, ww, cc)).astype(np.float32, order='C')
                                            biases = (np.ones((nn * ho * wo, kk)) * 10).astype(np.float32, order='C')

                                            w_c = weights.reshape(-1, kk)
                                            im2col_mm_result = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
                                            im2row_nhwc_cython(x, x_c,
                                                               kh, kw, ho, wo,
                                                               vpadding, hpadding, vstride, hstride,
                                                               vdilation, hdilation) @ w_c \
                                                + biases
                                            im2col_mm_result = im2col_mm_result.reshape(-1, ho, wo, kk)
                                            im2col_t = timeit(
                                                lambda: time_it_func(x, w_c, biases, b, kn, ho, wo, kh, kw, vpadding, hpadding, vstride, hstride, vdilation, hdilation),
                                                number=10) / 10

                                            conv_winograd_result = conv_winograd.conv_winograd_nhwc(weights, x, biases_wg,
                                                                                                    vpadding=vpadding,
                                                                                                    hpadding=hpadding,
                                                                                                    vstride=vstride,
                                                                                                    hstride=hstride,
                                                                                                    vdilation=vdilation,
                                                                                                    hdilation=hdilation)
                                            conv_winograd_t = timeit(
                                                lambda: conv_winograd.conv_winograd_nhwc(weights, x, biases_wg,
                                                                                         vpadding=vpadding,
                                                                                         hpadding=hpadding,
                                                                                         vstride=vstride, hstride=hstride,
                                                                                         vdilation=vdilation,
                                                                                         hdilation=hdilation),
                                                number=10) / 10
                                        case tensor_fmt:
                                            raise NotImplementedError(f"Unsupported tensor format {tensor_fmt}!")

                                    print(" conv_winograd time: {:.4f} ".format(conv_winograd_t), end="")
                                    print("mm time: {:.4f} ".format(im2col_t), end="")
                                    print("np.allclose:",
                                          np.allclose(conv_winograd_result, im2col_mm_result, atol=1e-3), end="")
                                    # print(" np.sum:", np.max(np.abs(conv_winograd_result-im2col_mm_result)), end="")
                                    print((" WINOGR", " IM2COL")[conv_winograd_t > im2col_t],
                                          im2col_t / conv_winograd_t)
    # """


if __name__ == "__main__":
    __usage_example__()



import numpy as np

from pydtnn.backends.cpu.layers.abstract.block_layer import AbstractBlockLayerCPU
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum


class ConcatenationBlockCPU(ConcatenationBlock, AbstractBlockLayerCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The next attributes will be initialized later
        self.out_co: list[int] = None  # type: ignore
        self.idx_co: np.ndarray = None  # type: ignore
        self.concat_dim: int = None  # type: ignore
    # ---

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_REPLICATE)
        _x:list[np.ndarray] = [np.zeros((0,))] * len(self.paths)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        for i, p in enumerate(self.paths):
            x_forward = x
            for layer in p:
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
                x_forward = layer.forward(x_forward)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            _x[i] = x_forward
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONCAT)
        y = np.concatenate(_x, axis=self.concat_dim)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def backward(self, dy: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SPLIT)
        dx: list[np.ndarray] = np.split(dy, self.idx_co[:-1], axis=self.concat_dim)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        num_paths = len(self.paths)

        p = self.paths[0]
        for layer in reversed(p):
            self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.BACKWARD)
            dx[0] = layer.backward(dx[0])
            self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

        for i in range(1, num_paths):
            p = self.paths[i]
            for layer in reversed(p):
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.BACKWARD)
                dx[i] = layer.backward(dx[i])
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_ELTW_SUM)
            np.add(dx[0], dx[i], out=dx[0],
                   dtype=self.model.dtype, order="C")
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx[0], dtype=self.model.dtype, order='C', copy=None)



import typing
from enum import auto, StrEnum

import numpy as np

type ArrayShape = tuple[int, ...]
DTYPE2CTYPE: dict[np.dtype, str] = {
    np.dtype(np.float32): "float",
    np.dtype(np.float64): "double"
}


class NetworkAlgEnum(StrEnum):
    BTA = auto()
    VDG = auto()

class Components(StrEnum):
    DATASETS = auto()
    ACTIVATIONS = auto()
    LAYERS = auto()
    LOSSES = auto()
    METRICS = auto()
    MODELS = auto()
    OPTIMIZERS = auto()
    SCHEDULERS = auto()
    TRACERS = auto()

class Parameters(StrEnum):
    PATHS = auto()
    CANONICAL_NAME = auto()
    MODEL_NAME = auto()
    LAYERS = auto()

    RUNNING_MEAN = auto()
    RUNNING_VAR = auto()
    GAMMA = auto()
    DGAMMA = auto()
    BETA = auto()
    DBETA = auto()

    WEIGHTS = auto()
    DW = auto()

    BIASES = auto()
    DB = auto()



# NOTE: It is necessary to have "ArrayShape" initialized before TensorGPU
if typing.TYPE_CHECKING:
    from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
type Array = "np.ndarray | TensorGPU"



from pydtnn.backends import PromoteToBackend
from pydtnn.utils import find_component
from pydtnn.utils.constants import Array
from pydtnn.utils.constants import ArrayShape


class Loss[T: Array](PromoteToBackend):
    format = ""

    def __init__(self, shape: ArrayShape, eps=1e-8):
        self.shape = shape
        self.eps = eps

    def initialize(self) -> None:
        pass

    def compute(self, y_pred: T, y_targ: T, batch_size: int) -> tuple[float, T]:
        raise NotImplementedError()


def select(name: str) -> type[Loss]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



from __future__ import annotations

from enum import auto, StrEnum

import numpy as np


# Formats
class ChannelFormat(StrEnum):
    HW = auto()
    WH = auto()

    def as_sample(self) -> SampleFormat:
        """Up-cast format to include channel number (prefers left-side up-cast)"""
        try:
            return SampleFormat(f"c{self}")
        except ValueError:
            return SampleFormat(f"{self}c")


class SampleFormat(StrEnum):
    CHW = auto()
    HWC = auto()
    WHC = auto()

    def as_channel(self) -> ChannelFormat:
        """Down-cast format to just the channel"""
        return ChannelFormat(self.strip("c"))

    def as_tensor(self) -> TensorFormat:
        """Up-cast format to include sample number (prefers left-side up-cast)"""
        try:
            return TensorFormat(f"n{self}")
        except ValueError:
            return TensorFormat(f"{self}n")


class TensorFormat(StrEnum):
    NCHW = auto()
    NHWC = auto()

    def as_sample(self) -> SampleFormat:
        """Down-cast format to just the sample"""
        return SampleFormat(self.strip("n"))


# Helpers
def format_reshape(shape: tuple[int, ...], src: str, dst: str) -> tuple[int, ...]:
    """
    Transform the `shape` from its current `src` order to `dst` order.

    Args:
        shape (tuple[int, ...]): source shape data.
        src (str): current format.
        dst (str): desired format.
    Returns:
        (tuple[int, ...]): `shape` with `dst` order.
    """

    assert len(shape) == len(src) == len(dst), f"Inconsistent number of dimensions ({shape=}, {src=}, {dst=})"
    assert set(src) == set(dst), f"Inconsistent dimension names ({src=}, {dst=})"
    assert len(src) == len(set(src)), f"Duplicate dimension names ({src=})"
    assert len(dst) == len(set(dst)), f"Duplicate dimension names ({dst=})"

    if src == dst:
        return shape
    dims = dict(zip(src, shape))
    return tuple(dims[i] for i in dst)  # type: ignore


def format_transpose(data: np.ndarray, src: str, dst: str) -> np.ndarray:
    """
    Transpose elements of `data` from its current `src` format to `dst` format.

    Args:
        data (np.ndarray): source numpy array.
        src (str): current format.
        dst (str): desired format.
    Returns:
        np.ndarray: `data` with `dst` format.
    """
    if src == dst:
        return data
    return data.transpose(format_reshape(range(len(data.shape)), src, dst))  # type: ignore


# Encoders
def encode_shape(shape: tuple[int, ...], format: str = TensorFormat.NCHW) -> tuple[int, ...]:
    """
    Transform the `shape` from its current `NCHW` order to `format` order (or less).

    Args:
        shape (tuple[int, ...]): source shape data.
        format (str): encoded format (NCHW).
    Returns:
        (tuple[int, ...]): `shape` with encoded order.
    """
    return format_reshape(shape, TensorFormat.NCHW[-len(shape):], format[-len(shape):])


def decode_shape(shape: tuple[int, ...], format: str = TensorFormat.NCHW) -> tuple[int, ...]:
    """
    Transform the `shape` from its current `format` order to `NCHW` order (or less).

    Args:
        shape (tuple[int, ...]): encoded shape data.
        format (str): encoded format (NCHW).
    Returns:
        (tuple[int, ...]): `shape` with `NCHW` order.
    """
    return format_reshape(shape, format[-len(shape):], TensorFormat.NCHW[-len(shape):])


def encode_tensor(data: np.ndarray, format: str = TensorFormat.NCHW) -> np.ndarray:
    """
    Transpose elements of `data` from its current `NCHW` format to `format` format.

    Args:
        data (np.ndarray): source data.
        format (str): encoded format (NCHW).
    Returns:
        (np.ndarray): `data` with encoded order.
    """
    return format_transpose(data, TensorFormat.NCHW[-len(data.shape):], format[-len(data.shape):])


def decode_tensor(data: np.ndarray, format: str = TensorFormat.NCHW) -> np.ndarray:
    """
    Transpose elements of `data` from its current `format` format to `NCHW` format.

    Args:
        data (np.ndarray): encoded data.
        format (str): encoded format (NCHW).
    Returns:
        (np.ndarray): `data` with `NCHW` order.
    """
    return format_transpose(data, format[-len(data.shape):], TensorFormat.NCHW[-len(data.shape):])



from pydtnn.layers.flatten import Flatten
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU


class FlattenGPU(Flatten[TensorGPU], LayerGPU):

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        self.y = x  # type: ignore (it's okay)

    def forward(self, x: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_Y)
        self.y.ary = x.ary.reshape((self.model.batch_size, *self.shape))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_RESHAPE_DX)
        self.dx = dy.reshape((self.model.batch_size, *self.prev_shape))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



import numpy as np

from pydtnn.backends.cpu.optimizers.optimizer import OptimizerCPU
from pydtnn.optimizers.sgd import SGD

from pydtnn.backends.cpu.layers.layer import LayerCPU


class SGDCPU(SGD[np.ndarray], OptimizerCPU):

    def initialize(self, list_layers: list[LayerCPU]) -> None:

        for layer in list_layers:
            list_grad_vars = list(layer.grad_vars.keys())
            if len(list_grad_vars) != 0:
                self.context[layer.id] = dict[str, np.ndarray]()  # type: ignore 
                for w_ in list_grad_vars:
                    w: np.ndarray = getattr(layer, w_)
                    self.context[layer.id]["velocity_%s" % w_] = np.zeros_like(w, dtype=layer.model.dtype, order="C")

    def update(self, layer: LayerCPU) -> None:
        for w_, dw_ in layer.grad_vars.items():
            w, dw = getattr(layer, w_), getattr(layer, dw_)
            velocity: np.ndarray = self.context[layer.id]["velocity_%s" % w_]  # type: ignore 
            w: np.ndarray
            dw: np.ndarray

            if not (self.are_all_zeros(velocity) and self.are_all_zeros(w) and self.are_all_zeros(dw)):
                # NOTE: The operations are unrolled in order to reduce the memory consumed by intermediate copies of the variables during the operations.

                # velocity = self.momentum * velocity + dw
                
                np.multiply(velocity, self.momentum, out=velocity, 
                            dtype=self.dtype)
                np.add(velocity, dw, out=velocity, 
                       dtype=self.dtype)

                # if self.nesterov:
                #    w -= self.learning_rate * (self.decay * w + dw + self.momentum * velocity)
                # else:
                #    w -= self.learning_rate * (self.decay * w + velocity)
                if self.nesterov:
                    v = np.multiply(velocity, self.momentum, dtype=self.dtype, order="C")
                    np.add(v, dw, out=v, 
                           dtype=self.dtype)
                else:
                    v = velocity
                temp_w = np.multiply(w, self.decay, dtype=self.dtype, order="C")
                np.add(temp_w, v, out=temp_w,
                       dtype=self.dtype)
                np.multiply(temp_w, self.learning_rate, out=temp_w, 
                            dtype=self.dtype)
                np.subtract(w, temp_w, out=w,
                            dtype=self.dtype)
                del temp_w
            # else: continue



import unittest

from pydtnn.activations.relu import Relu
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.batch_normalization_relu import BatchNormalizationRelu
from pydtnn.model import Model
from pydtnn.tests.abstract.common import D, Params
from pydtnn.tests.abstract.conv_2d_common import Conv2DCommonTestCase
from pydtnn.utils.tensor import TensorFormat


class BatchNormalizationReluTestCase(Conv2DCommonTestCase):
    """
    Tests that BatchNormalization+Relu leads to the same results than BatchNormalizationRelu
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global Conv2DCommonTestCase
    del Conv2DCommonTestCase

    @staticmethod
    def _get_layers(d: D, deconv=False, trans=False) -> tuple:
        params = Params()
        params.tensor_format = TensorFormat.NCHW.upper()
        params.batch_size = d.b
        params.conv_variant = "gemm"
        model = Model(**vars(params))
        model.mode = Model.Mode.TRAIN

        bn = BatchNormalization()
        relu = Relu()
        chain = ConcatenationBlock([
            bn,
            relu
        ])
        shape = (d.c, d.h, d.w)
        chain.init_backend_from_model(model)
        chain.initialize(prev_shape=shape, x=None)

        from_parent = relu.__dict__ | bn.__dict__
        fuse = BatchNormalizationRelu(from_parent=from_parent)
        fuse.init_backend_from_model(model)
        fuse.__dict__.update(from_parent)
        fuse.initialize(prev_shape=shape, x=None)

        # Set the same initial weights and biases to both layers
        fuse.running_mean = bn.running_mean.copy()
        fuse.running_var = bn.running_var.copy()

        return chain, fuse

    @staticmethod
    def _set_state(layer: BatchNormalization, weights) -> None:
        pass

    @unittest.skip("Backward not implemented")
    def test_forward_backward_larger_handmade_array_stride3(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_larger_handmade_array_stride2(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_handmade_array(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride3_filter1x2(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride3_filter1x1(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_defaults(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_handmade_array_stride2(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride3(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride12(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_alexnet_imagenet_first_conv2d(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_alexnet_cifar10_first_conv2d(self):
        raise NotImplementedError()



import numpy as np

from pydtnn.backends.cpu.losses.loss import LossCPU
from pydtnn.losses.categorical_cross_entropy import CategoricalCrossEntropy


class CategoricalCrossEntropyCPU(CategoricalCrossEntropy[np.ndarray], LossCPU):

    def compute(self, _y_pred: np.ndarray, y_targ: np.ndarray, batch_size: int) -> tuple[float, np.ndarray]:
        # Common
        y_pred: np.ndarray = np.clip(_y_pred, a_min=self.eps, a_max=(1 - self.eps))
        b_range: np.ndarray = np.arange(y_pred.shape[0])

        # Loss
        loss: float = -np.sum(np.log(y_pred[b_range, np.argmax(y_targ, axis=1)])) / y_pred.shape[0]

        # DX
        # NOTE/FIXME: The last line before the return raise an error if the model works in int8 due it is trying to store an float64 into a int8 [possible fix below].
        dx: np.ndarray = np.copy(y_targ)
        # dx:np.ndarray = y_targ.astype(np.float64, copy=True)
        dx_amax: np.ndarray = np.argmax(dx, axis=1)
        # NOTE/FIXME: This will raise an error if the model works in int8 due it is trying to store an float64 into a int8.
        dx[b_range, dx_amax] /= (-y_pred[b_range, dx_amax] * batch_size)

        return loss, np.asarray(dx, dtype=self.model.dtype, order="C", copy=None)



"""
PyDTNN initializers
"""

import numpy as np
import scipy.stats as stats
from enum import StrEnum, auto
from typing import Callable
from pydtnn.utils.constants import ArrayShape
from pydtnn.utils import random


class DistributionModeEnum(StrEnum):
    FAN_IN = auto()
    FAN_OUT = auto()
    FAN_AVG = auto()
# ---


class ProbabilisticDistribution(StrEnum):
    UNIFORM = auto()
    NORMAL = auto()
# ---


# 0.879... = scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
STD_DEV_CONST = 0.87962566103423978


def _compute_fans(shape: ArrayShape) -> tuple[int, int]:
    if len(shape) == 2:
        fan_in, fan_out = shape[0], shape[1]
    elif len(shape) > 2:
        receptive_field = np.prod(shape[2:])
        fan_in = shape[1] * receptive_field
        fan_out = shape[0] * receptive_field
    else:
        raise ValueError(f"The length of 'shape' must be greater or equal to 2, it is {len(shape)}.")
    return fan_in, fan_out


def _generate_distribution(shape: ArrayShape, scale: float, mode: DistributionModeEnum,
                           distribution: ProbabilisticDistribution, dtype: np.dtype) -> np.ndarray:
    fan_in, fan_out = _compute_fans(shape)

    match mode:
        case DistributionModeEnum.FAN_IN:
            scale /= max(1., fan_in)
        case DistributionModeEnum.FAN_OUT:
            scale /= max(1., fan_out)
        case DistributionModeEnum.FAN_AVG:
            scale /= max(1., float(fan_in + fan_out) / 2)
        case _:
            raise NotImplementedError(f"mode: \'{mode}\' not implemented")

    match distribution:
        case ProbabilisticDistribution.NORMAL:
            stddev: float = np.sqrt(scale) / STD_DEV_CONST
            # Truncated normal distribution [-2*stddev, 2*stddev]
            x = stats.truncnorm(-2 * stddev, 2 * stddev, loc=0, scale=stddev).rvs(shape).astype(dtype, copy=False, order="C")
        case ProbabilisticDistribution.UNIFORM:
            limit = np.sqrt(3. * scale)
            x = random.uniform(-limit, limit, shape).astype(dtype, copy=False, order="C")
        case _:
            raise NotImplementedError(f"distribution: \'{distribution}\' not implemented")
    return x


def glorot_uniform(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return _generate_distribution(shape, 1.0, DistributionModeEnum.FAN_AVG,
                                  ProbabilisticDistribution.UNIFORM, dtype)


def glorot_normal(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return _generate_distribution(shape, 1.0, DistributionModeEnum.FAN_AVG, ProbabilisticDistribution.NORMAL, dtype)


def he_uniform(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return _generate_distribution(shape, 2.0, DistributionModeEnum.FAN_IN, ProbabilisticDistribution.UNIFORM, dtype)


def he_normal(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return _generate_distribution(shape, 2.0, DistributionModeEnum.FAN_IN, ProbabilisticDistribution.NORMAL, dtype)


def lecun_uniform(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return _generate_distribution(shape, 1.0, DistributionModeEnum.FAN_IN, ProbabilisticDistribution.UNIFORM, dtype)


def lecun_normal(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return _generate_distribution(shape, 1.0, DistributionModeEnum.FAN_IN, ProbabilisticDistribution.NORMAL, dtype)


def ones(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return np.ones(shape, dtype=dtype, order="C")


def zeros(shape: ArrayShape, dtype: np.dtype) -> np.ndarray:
    return np.zeros(shape, dtype=dtype, order="C")


type InitializerFunc = Callable[[ArrayShape, np.dtype], np.ndarray]



from pydtnn.metrics.confusion_matrix import ConfusionMatrix
from pydtnn.utils.constants import Array


class BinaryConfusionMatrix[T: Array](ConfusionMatrix[T]):
    conf_matrix: T = None  # type: ignore



import numpy as np

from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.scalar import Scalar


class ScalarCPU(Scalar[np.ndarray], LayerCPU):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        # Performance model
        self.fwd_time = None  # Not yet
        self.bwd_time = self.fwd_time

    def forward(self, x):
        return x * self.scale

    def backward(self, dy):
        return dy * self.scale



import os
import itertools
from typing import TYPE_CHECKING

import numpy as np

from pydtnn.datasets.dataset import Dataset
from pydtnn.utils.tensor import TensorFormat


if TYPE_CHECKING:
    from pydtnn.model import Model

TRAIN_NSAMPLES = 60000
TEST_NSAMPLES = 10000
INPUT_SHAPE = (1, 28, 28)
OUTPUT_SHAPE = (10,)


class MNIST(Dataset):
    """
    MNIST Dataset

    Handwritten digit database.

    Source (SHA1):
    6c95f4b05d2bf285e1bfb0e7960c31bd3b3f8a7d train-images-idx3-ubyte.gz https://ossci-datasets.s3.amazonaws.com/mnist/train-images-idx3-ubyte.gz
    2a80914081dc54586dbdf242f9805a6b8d2a15fc train-labels-idx1-ubyte.gz https://ossci-datasets.s3.amazonaws.com/mnist/train-labels-idx1-ubyte.gz
    c3a25af1f52dad7f726cce8cacb138654b760d48 t10k-images-idx3-ubyte.gz https://ossci-datasets.s3.amazonaws.com/mnist/t10k-images-idx3-ubyte.gz
    763e7fa3757d93b0cdec073cef058b2004252c17 t10k-labels-idx1-ubyte.gz https://ossci-datasets.s3.amazonaws.com/mnist/t10k-labels-idx1-ubyte.gz

    Normalize (z-score):
    offset: -0.131
    scale:  +3.237
    """

    def __init__(self, model: "Model", force_test_as_validation=False, debug=False):
        super().__init__(model, TRAIN_NSAMPLES, TEST_NSAMPLES, INPUT_SHAPE, OUTPUT_SHAPE, force_test_as_validation=force_test_as_validation, debug=debug)

    def _init_actual_data(self) -> None:
        self._x_filename = [
            os.path.join(self.model.dataset_path, "train-images-idx3-ubyte.gz"),
            None,
            os.path.join(self.model.dataset_path, "t10k-images-idx3-ubyte.gz")
        ]
        self._y_filename = [
            os.path.join(self.model.dataset_path, "train-labels-idx1-ubyte.gz"),
            None,
            os.path.join(self.model.dataset_path, "t10k-labels-idx1-ubyte.gz")
        ]
        if self.test_as_validation:
            self._x_filename[Dataset.Part.VAL] = self._x_filename[Dataset.Part.TEST]
            self._y_filename[Dataset.Part.VAL] = self._y_filename[Dataset.Part.TEST]
        else:
            self._x_filename[Dataset.Part.VAL] = self._x_filename[Dataset.Part.TRAIN]
            self._y_filename[Dataset.Part.VAL] = self._y_filename[Dataset.Part.TRAIN]

        self._images_header_offset = 16  # 4 + 4 * 3
        self._labels_header_offset = 8  # 4 + 4 * 1

        # Pregenerate GZIP indexs
        for gz in itertools.chain(self._x_filename, self._y_filename):
            self._gzip_open(gz).close()

    def _actual_data_generator(self, part: Dataset.Part):
        size = int(np.prod(INPUT_SHAPE))
        offset = self._images_header_offset + self._local_offset[part] * size
        nbytes = self._local_nsamples[part] * size
        filename = self._x_filename[part]
        with self._gzip_open(filename) as f:
            x = self._read_file(f, offset, nbytes).reshape(self._local_nsamples[part], *INPUT_SHAPE) / 255.0
        x = x.astype(self.model.dtype)

        x = self.model.encode_tensor(x)

        offset = self._labels_header_offset + self._local_offset[part] * 1  # The output class is encoded as a number
        nbytes = self._local_nsamples[part] * 1  # The output class is encoded as a number

        with self._gzip_open(self._y_filename[part]) as f:
            y_classes = self._read_file(f, offset, nbytes)

        y = np.zeros((self._local_nsamples[part], *self.output_shape), dtype=self.model.dtype, order="C")
        self._decode_class(y, y_classes)

        yield x, y

    def _read_file(self, f, offset: int, nbytes: int) -> np.ndarray:
        # How to read the header:
        #  zero, data_type, dims = struct.unpack('>HBB', f.read(4))
        #  shape = (struct.unpack('>I', f.read(4))[0] for _ in range(dims))
        f.seek(offset)
        return np.frombuffer(f.read(nbytes), dtype=np.uint8)



from pydtnn.backends.gpu.layers.abstract.block_layer import AbstractBlockLayerGPU
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum
from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU


class AdditionBlockGPU(AdditionBlock[TensorGPU], AbstractBlockLayerGPU):
    y: TensorGPU
    def forward(self, x: TensorGPU) -> TensorGPU:
        for i, p in enumerate(self.paths):
            y_i = x
            for layer in p:
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
                y_i = layer.forward(y_i)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            if i == 0:
                self.y = y_i
            else:
                alpha, beta = 1.0, 1.0
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                             self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_ELTW_SUM)
                # noinspection PyUnboundLocalVariable
                cudnn.cudnnAddTensor(self.model.cudnn_handle, alpha, y_i.desc,
                                     y_i.ptr, beta, self.y.desc, self.y.ptr)
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        for i, p in enumerate(self.paths):
            dx_i = dy
            for layer in reversed(p):
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.BACKWARD)
                dx_i = layer.backward(dx_i)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            if i == 0:
                self.dx = dx_i
            else:
                alpha, beta = 1.0, 1.0
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                             self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_ELTW_SUM)
                # noinspection PyUnboundLocalVariable
                cudnn.cudnnAddTensor(self.model.cudnn_handle, alpha, dx_i.desc,
                                     dx_i.ptr, beta, self.dx.desc, self.dx.ptr)
                self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



from pydtnn.activations.relu import Relu
from pydtnn.utils.constants import ArrayShape, Array

from typing import Self
class LeakyRelu[T: Array](Relu[T]):

    def __init__(self, shape: ArrayShape = (1,), negative_slope: float = 0.01):
        super().__init__(shape)
        self.negative_slope: float = negative_slope



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform


def resnet18(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    _(Conv2D(nfilters=64, filter_shape=(3, 3), stride=1, padding=1, weights_initializer=he_uniform))
    _(BatchNormalization())

    layout = [[64, 2, 1], [128, 2, 2], [256, 2, 2], [512, 2, 2]]  # Resnet-18
    for n_filt, res_blocks, stride in layout:
        for r in range(res_blocks):
            if r > 0:
                stride = 1
            _(AdditionBlock(
                [
                    Conv2D(nfilters=n_filt, filter_shape=(3, 3), stride=stride, padding=1,
                           weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu(),
                    Conv2D(nfilters=n_filt, filter_shape=(3, 3), stride=1, padding=1,
                           weights_initializer=he_uniform),
                    BatchNormalization()
                ],
                [
                    Conv2D(nfilters=n_filt, filter_shape=(1, 1), stride=stride, weights_initializer=he_uniform),
                    BatchNormalization()
                ] if stride != 1 else []))
            _(Relu())

    _(AveragePool2D(pool_shape=(0, 0)))  # Global average pooling 2D
    _(Flatten())
    _(FC(shape=(512,)))
    _(BatchNormalization())
    _(Relu())
    _(FC(shape=output_shape, activation=Softmax))

    return model



from typing import TYPE_CHECKING
from warnings import warn

from pydtnn.layer_base import LayerBase
from pydtnn.optimizers.optimizer import Optimizer
from pydtnn.utils.constants import Array

if TYPE_CHECKING:
    from pydtnn.model import Model


class OkTopk[T: Array](Optimizer[T]):
    """
    SGD Ok-Topk Optimizer
    """

    def __init__(self, learning_rate: float = 1e-2, momentum: float = 0.9,
                 tau: int = 64, tau_prime: int = 32, density: float = 0.01, min_k_layer: int = 10):

        super().__init__()
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.residuals = {}
        self.tau = tau
        self.tau_prime = tau_prime
        self.density = density
        self.min_k_layer = min_k_layer
        self.iterations = {}
        self.all_local_th = {}
        self.all_global_th = {}
        self.all_residuals = {}
        self.all_boundaries = {}
        self.info_messages = set()

    def initialize(self, list_layers: list[LayerBase]) -> None:
        if self.model.model_sync_freq >= 0:
            warn("Optimizer does model sync but global model sync is also enabled!", RuntimeWarning)

        if not self.model.shared_storage:
            raise NotImplementedError("OkTopK optimizer does not support Federated Learing (unbalanced datasets)!")

    @classmethod
    def from_model(cls, model: "Model") -> "OkTopk":
        return OkTopk(learning_rate=model.learning_rate,
                      momentum=model.optimizer_momentum,
                      tau=model.optimizer_tau,
                      tau_prime=model.optimizer_tau_prime,
                      density=model.optimizer_density)



from typing import Any
from pydtnn.layers.conv_2d import Conv2D

import pycuda.driver as drv  #type: ignore
import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.compiler import SourceModule  #type: ignore
from pycuda.driver import Function  #type: ignore

import numpy as np

from pydtnn.utils.performance_models import matmul_time
from pydtnn.backends.gpu.layers.layer import LayerGPU

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import ArrayShape, DTYPE2CTYPE, Parameters

class Conv2DGPU(Conv2D[TensorGPU], LayerGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The following attributes will be initalized later.
        self.fwd_algo: int = None  #type: ignore
        self.bwd_dw_algo: int = None  #type: ignore
        self.bwd_dx_algo: int = None  #type: ignore
        self.conv_desc = None
    # ----

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.stream_2 = drv.Stream()

        self.weights_cpu = self.weights_initializer(self.weights_shape, self.model.dtype)
        weights_gpu = gpuarray.to_gpu(self.weights_cpu)
        self.weights = TensorGPU(weights_gpu, self.model.tensor_format, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.FILTER)
        # Biases
        if self.use_bias:
            biases_shape = self.model.encode_shape((1, self.co, 1, 1))
            self.biases_cpu = self.biases_initializer(biases_shape, self.model.dtype)
            biases_gpu = gpuarray.to_gpu(self.biases_cpu)
            self.biases = TensorGPU(biases_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.fwd_time = \
            matmul_time(m=self.co, n=(self.model.batch_size * self.ho * self.wo), k=(self.ci * self.kh * self.kw),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw, dtype=self.model.dtype) 
        self.bwd_time = \
            matmul_time(m=self.co, n=(self.ci * self.kh * self.kw), k=(self.model.batch_size * self.ho * self.wo),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw, dtype=self.model.dtype) + \
            matmul_time(m=(self.ci * self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo), k=self.co,
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw, dtype=self.model.dtype)  # type: ignore (It is correct.)

        if self.model.gpudirect:
            bias_tensor_type = TensorGPU.TensorTypeEnum.FILTER
            _drv = drv
        else:
            bias_tensor_type = TensorGPU.TensorTypeEnum.TENSOR
            _drv = None

        # Derivative dw and derivative db
        self.dw_cpu, self.dw = TensorGPU.initialize(self.weights.ary.shape, self.model.dtype, tensor_format=self.model.tensor_format,
                                                    cudnn_dtype=self.model.cudnn_dtype, gpudirect=self.model.gpudirect, 
                                                    tensor_type=TensorGPU.TensorTypeEnum.FILTER, drv=_drv)
        if self.use_bias:
            self.biases: TensorGPU
            self.db_cpu, self.db = TensorGPU.initialize(self.biases.ary.shape, self.model.dtype, tensor_format=self.model.tensor_format,
                                                        cudnn_dtype=self.model.cudnn_dtype, gpudirect=self.model.gpudirect, 
                                                        tensor_type=bias_tensor_type, drv=_drv)
    # ----

    def _export_weights_dw(self, key: str) -> Any:
        # NOTE: Every variant must implement their version of this method.
        #super()._export_prop(key)
        msg = "This is a \"fake\" function. It must be overrided by the child classes."
        raise NotImplementedError(f"Conv2DGPU forward: {msg}")
    # ----

    def _export_biases_db(self, key: str) -> Any:
        value = getattr(self, key)
        gpu_ary = value.ary
        cpu_ary = gpu_ary.get()

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                return np.asarray(np.squeeze(cpu_ary, axis=(0, 1, 2)), dtype=np.float64, order="C", copy=True)
            case TensorFormat.NCHW:
                return np.asarray(np.squeeze(cpu_ary, axis=(0, 2, 3)), dtype=np.float64, order="C", copy=True)
            case _:
                return super()._export_prop(key)
    # ----

    def _export_prop(self, key: str) -> Any:
        match key:
            case Parameters.WEIGHTS | Parameters.DW:
                return self._export_weights_dw(key)
            case Parameters.BIASES | Parameters.DB:
                return self._export_biases_db(key)
            case _:
                return super()._export_prop(key)
    # ----

    def _import_biases_db(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        
        match self.model.tensor_format:
            case TensorFormat.NHWC:
                cpu_ary = np.asarray(np.expand_dims(value, axis=(0, 1, 2)), dtype=self.model.dtype, order="C", copy=None)
                attribute.ary.set(cpu_ary)
                return
            case TensorFormat.NCHW:
                cpu_ary = np.asarray(np.expand_dims(value, axis=(0, 2, 3)), dtype=self.model.dtype, order="C", copy=None)
                attribute.ary.set(cpu_ary)
                return
            # case default: (next return)
        return super()._import_prop(key, value)
    # ----

    def _import_weights_dw(self, key: str, value: Any) -> None:
        # NOTE: Every variant must implement their version of this method.
        #super()._export_prop(key)
        msg = "This is a \"fake\" function. It must be overrided by the child classes"
        raise NotImplementedError(f"Conv2DGPU forward: {msg}")
    # ----

    def _import_prop(self, key: str, value) -> None:
        match key:
            case Parameters.WEIGHTS | Parameters.DW:
                return self._import_weights_dw(key, value)
            case Parameters.BIASES | Parameters.DB:
                return self._import_biases_db(key, value)
            # 
            case _:
                return super()._import_prop(key, value)
    # ----

    def forward(self, x: TensorGPU) -> TensorGPU:
        msg = "This is a fake forward function. It must be masked on initialization by a _forward implementation."
        raise NotImplementedError(f"Conv2DGPU forward: {msg}")

    def backward(self, dy: TensorGPU) -> TensorGPU:
        msg = "This is a fake backward function. It must be masked on initialization by a _backward implementation."
        raise NotImplementedError(f"Conv2DGPU backward: {msg}")


#########################################################################################################
## CUDA-RELATED COMMON CODE ##
##############################
    def cuda_sum_bias_axis_023(self, _func_name: str = "bias_sum_bwd_depthwise_conv_nchw") -> Function:
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        # np.sum(dy, axis=(0, 2, 3), out=self.db)
        code = \
            """
__global__ void {func_name}({T}* dy, {T}* db
                            int c, int h, int w,
                            int N, int num_workers)
{{
    int idx, index_c;

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        index_c = (idx / (h*w)) % c;
        *(db + index_c) += *(dy + idx);
    }}
}}
"""

        code = code.format(func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ----

    def cuda_sum_bias_axis_012(self, _func_name: str = "bias_sum_bwd_depthwise_conv_nhwc") -> Function:
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        # np.sum(dy, axis=(0, 1, 2), out=self.db)
        code = \
            """
__global__ void {func_name}({T}* dy, {T}* db,
                            int c, int N,
                            int num_workers)
{{
    int idx;

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        *(db + (idx % c)) += *(dy + idx);
    }}
}}
"""

        code = code.format(func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ----

# CUDA-related constans

MACROS_NCHW = \
    """
#define SHIFT_POINTER(p, c, h, w, ni, ci, hi, wi) p + ((ni * c + ci) * h + hi) * w + wi
#define SHIFT_POINTER_K(p, c, yc, ci, yci) p + (yci * c + ci)
#define INDEX_N(idx, N, n) idx * n / N
#define INDEX_C(idx, c, h, w) (idx / (h * w)) % c
#define INDEX_H(idx, c, h, w) (idx / w) % h
#define INDEX_W(idx, c, h, w) idx % w
"""
# ---

MACROS_NHWC = \
    """
#define SHIFT_POINTER(p, c, h, w, ni, ci, hi, wi) p + ((ni * h + hi) * w + wi) * c + ci
#define SHIFT_POINTER_K(p, c, yc, ci, yci) p + (ci * yc + yci)
#define INDEX_N(idx, N, n) idx * n / N
#define INDEX_H(idx, h, w, c) (idx / (w * c)) % h
#define INDEX_W(idx, h, w, c) (idx / c) % w
#define INDEX_C(idx, h, w, c) idx % c
"""
# ---
#########################################################################################################



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydtnn.model import Model
from pydtnn.layers.layer import Layer
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum

try:
    from pydtnn.libs.libmpi import MPI
except Exception:
    pass
from pydtnn.utils.constants import ArrayShape
import numpy as np


class LayerCPU(Layer[np.ndarray]):
    """
    Extends a Layer class with the attributes and methods required by CPU Layers.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model: Model[np.ndarray]

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)

    @property
    def _ary_prop(self) -> set[str]:
        return {*self.grad_vars.keys(), *self.grad_vars.values()}

    def _export_prop(self, key: str):
        if key not in self._ary_prop:
            return super()._export_prop(key)

        ary = getattr(self, key)
        return np.asarray(ary, dtype=np.float64, order="C", copy=True)

    def _import_prop(self, key: str, value) -> None:
        if key not in self._ary_prop:
            return super()._import_prop(key, value)

        ary = getattr(self, key)
        ary[:] = np.asarray(value, dtype=self.model.dtype, order="C", copy=None)

    def reduce_weights_async(self, gradient=True):
        if not self.model.comm:
            return
        self.reqs_allred = {}

        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            dw: np.ndarray = getattr(self, dw_)
            dw *= self.model.rank_weight
            if self.model.crypt:
                dw = self.model.crypt.encrypt(dw)  # type: ignore
            if self.model.use_mpi_buffers:
                req = self.model.comm.Iallreduce(MPI.IN_PLACE, dw, op=MPI.SUM)
            else:
                req = self.model.comm.iallreduce(dw, op=MPI.SUM)
            self.reqs_allred[dw_] = req

    def wait_allreduce_async(self, gradient=True):
        if not self.model.comm or self.model.enable_nccl:
            return
        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            res = self.reqs_allred[dw_].wait()
            if res is None:
                dw = getattr(self, dw_)
            else:
                dw = res
            if self.model.crypt:
                dw = self.model.crypt.decrypt(dw)
            setattr(self, dw_, dw)

    def reduce_weights_sync(self, gradient=True):
        if not self.model.comm:
            return
        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            self.model.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT],
                                          [self.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.ALLREDUCE_DW,
                                           self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.OPS_ALLREDUCE_DW])
            dw: np.ndarray = getattr(self, dw_)
            dw *= self.model.rank_weight
            if self.model.crypt:
                dw = self.model.crypt.encrypt(dw)  # type: ignore
            if self.model.use_mpi_buffers:
                self.model.comm.Allreduce(MPI.IN_PLACE, dw, op=MPI.SUM)
            else:
                dw = self.model.comm.allreduce(dw, op=MPI.SUM)
            if self.model.crypt:
                dw = self.model.crypt.decrypt(dw)  # type: ignore
            setattr(self, dw_, dw)
            self.model.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT], [PYDTNN_EVENT_FINISHED, PYDTNN_EVENT_FINISHED])



import numpy as np

from pydtnn.backends.cpu.optimizers.optimizer import OptimizerCPU
from pydtnn.optimizers.adam import Adam
from pydtnn.backends.cpu.layers.layer import LayerCPU


class AdamCPU(Adam[np.ndarray], OptimizerCPU):

    def initialize(self, list_layers: list[LayerCPU]) -> None:

        for layer in list_layers:
            self.context[layer.id] = dict[str, int | np.ndarray]()
            self.context[layer.id]["it"] = 0

            for w_ in layer.grad_vars.keys():
                w: np.ndarray = getattr(layer, w_)
                self.context[layer.id]["m_%s" % w_] = np.zeros_like(w, dtype=layer.model.dtype, order="C")
                self.context[layer.id]["v_%s" % w_] = np.zeros_like(w, dtype=layer.model.dtype, order="C")

    def update(self, layer: LayerCPU) -> None:
        self.context[layer.id]["it"] += 1
        it: int = self.context[layer.id]["it"]  # type: ignore

        for w_, dw_ in layer.grad_vars.items():
            w, dw = getattr(layer, w_), getattr(layer, dw_)
            w: np.ndarray
            dw: np.ndarray
            # Momentum of the weight or bias of the given layer
            m: np.ndarray = self.context[layer.id]["m_%s" % w_]  # type: ignore 
            # Velocity of the weight or bias of the given layer
            v: np.ndarray = self.context[layer.id]["v_%s" % w_]  # type: ignore

            if not (self.are_all_zeros(w) and self.are_all_zeros(dw) and self.are_all_zeros(m) and self.are_all_zeros(v)):
                # NOTE: The operations are unrolled in order to reduce the memory consumed by intermediate copies of the variables during the operations.
                # m = self.beta1 * m + (1 - self.beta1) * dw
                inv_beta1 = (1 - self.beta1)
                inv_beta2 = (1 - self.beta2)

                temp_dw = np.multiply(inv_beta1, dw, dtype=self.dtype, order="C")

                np.multiply(m, self.beta1, out=m, 
                            dtype=self.dtype)
                np.add(m, temp_dw, out=m,
                       dtype=self.dtype)

                # v = self.beta2 * v + (1 - self.beta2) * dw ** 2
                temp_dw = np.pow(dw, 2, dtype=self.dtype, order="C")

                np.multiply(v, self.beta2, out=v,
                            dtype=self.dtype)
                np.multiply(temp_dw, inv_beta2, out=temp_dw, 
                            dtype=self.dtype)
                np.add(v, temp_dw, out=v, 
                        dtype=self.dtype)
                del temp_dw

                mt: np.ndarray = np.divide(m, (inv_beta1 ** it), dtype=self.dtype, order="C")
                vt: np.ndarray = np.divide(v, (inv_beta2 ** it), dtype=self.dtype, order="C")

                # w -= self.learning_rate * (self.decay * w + (mt / np.sqrt(vt + self.epsilon)))
                temp_w = np.multiply(self.decay, w, dtype=self.dtype, order="C")

                np.add(vt, self.epsilon, out=vt,
                       dtype=self.dtype)
                np.sqrt(vt, out=vt, 
                        dtype=self.dtype)
                np.divide(mt, vt, out=mt,
                          dtype=self.dtype)
                del vt

                np.add(temp_w, mt, out=temp_w, 
                       dtype=self.dtype)
                del mt
                np.multiply(temp_w, self.learning_rate, out=temp_w, 
                            dtype=self.dtype)

                np.subtract(w, temp_w, out=w, 
                            dtype=self.dtype, order="C")
                del temp_w
            # else: continue



from pydtnn.layer_base import LayerBase
from pydtnn.utils import find_component
from pydtnn.utils.constants import ArrayShape, Array

class Activation[T: Array](LayerBase):

    def __init__(self, shape: ArrayShape = (1,)):
        super().__init__(shape)
        self.y: T = None  #type: ignore (it will be initalized later)

    def initialize(self, prev_shape: ArrayShape, x: T | None = None):
        super().initialize(prev_shape, x)
        self.shape = prev_shape


def select(name: str) -> type[Activation]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



"""
PyDTNN: TF to PyDTNN ResNet-50-v1.5 weights converter
"""

import os
import sys
import h5py
import urllib.request
import numpy as np
from pydtnn.utils.constants import Parameters

# PyDTNN <-> Keras layer conversion
layers = {"1_Conv2D": "conv1_conv",
          "2_BatchNormalization": "conv1_bn",
          "6_Conv2D": "conv2_block1_1_conv",
          "7_BatchNormalization": "conv2_block1_1_bn",
          "9_Conv2D": "conv2_block1_2_conv",
          "10_BatchNormalization": "conv2_block1_2_bn",
          "12_Conv2D": "conv2_block1_3_conv",
          "13_BatchNormalization": "conv2_block1_3_bn",
          "14_Conv2D": "conv2_block1_0_conv",
          "15_BatchNormalization": "conv2_block1_0_bn",
          "18_Conv2D": "conv2_block2_1_conv",
          "19_BatchNormalization": "conv2_block2_1_bn",
          "21_Conv2D": "conv2_block2_2_conv",
          "22_BatchNormalization": "conv2_block2_2_bn",
          "24_Conv2D": "conv2_block2_3_conv",
          "25_BatchNormalization": "conv2_block2_3_bn",
          "28_Conv2D": "conv2_block3_1_conv",
          "29_BatchNormalization": "conv2_block3_1_bn",
          "31_Conv2D": "conv2_block3_2_conv",
          "32_BatchNormalization": "conv2_block3_2_bn",
          "34_Conv2D": "conv2_block3_3_conv",
          "35_BatchNormalization": "conv2_block3_3_bn",
          "38_Conv2D": "conv3_block1_1_conv",
          "39_BatchNormalization": "conv3_block1_1_bn",
          "41_Conv2D": "conv3_block1_2_conv",
          "42_BatchNormalization": "conv3_block1_2_bn",
          "44_Conv2D": "conv3_block1_3_conv",
          "45_BatchNormalization": "conv3_block1_3_bn",
          "46_Conv2D": "conv3_block1_0_conv",
          "47_BatchNormalization": "conv3_block1_0_bn",
          "50_Conv2D": "conv3_block2_1_conv",
          "51_BatchNormalization": "conv3_block2_1_bn",
          "53_Conv2D": "conv3_block2_2_conv",
          "54_BatchNormalization": "conv3_block2_2_bn",
          "56_Conv2D": "conv3_block2_3_conv",
          "57_BatchNormalization": "conv3_block2_3_bn",
          "60_Conv2D": "conv3_block3_1_conv",
          "61_BatchNormalization": "conv3_block3_1_bn",
          "63_Conv2D": "conv3_block3_2_conv",
          "64_BatchNormalization": "conv3_block3_2_bn",
          "66_Conv2D": "conv3_block3_3_conv",
          "67_BatchNormalization": "conv3_block3_3_bn",
          "70_Conv2D": "conv3_block4_1_conv",
          "71_BatchNormalization": "conv3_block4_1_bn",
          "73_Conv2D": "conv3_block4_2_conv",
          "74_BatchNormalization": "conv3_block4_2_bn",
          "76_Conv2D": "conv3_block4_3_conv",
          "77_BatchNormalization": "conv3_block4_3_bn",
          "80_Conv2D": "conv4_block1_1_conv",
          "81_BatchNormalization": "conv4_block1_1_bn",
          "83_Conv2D": "conv4_block1_2_conv",
          "84_BatchNormalization": "conv4_block1_2_bn",
          "86_Conv2D": "conv4_block1_3_conv",
          "87_BatchNormalization": "conv4_block1_3_bn",
          "88_Conv2D": "conv4_block1_0_conv",
          "89_BatchNormalization": "conv4_block1_0_bn",
          "92_Conv2D": "conv4_block2_1_conv",
          "93_BatchNormalization": "conv4_block2_1_bn",
          "95_Conv2D": "conv4_block2_2_conv",
          "96_BatchNormalization": "conv4_block2_2_bn",
          "98_Conv2D": "conv4_block2_3_conv",
          "99_BatchNormalization": "conv4_block2_3_bn",
          "102_Conv2D": "conv4_block3_1_conv",
          "103_BatchNormalization": "conv4_block3_1_bn",
          "105_Conv2D": "conv4_block3_2_conv",
          "106_BatchNormalization": "conv4_block3_2_bn",
          "108_Conv2D": "conv4_block3_3_conv",
          "109_BatchNormalization": "conv4_block3_3_bn",
          "112_Conv2D": "conv4_block4_1_conv",
          "113_BatchNormalization": "conv4_block4_1_bn",
          "115_Conv2D": "conv4_block4_2_conv",
          "116_BatchNormalization": "conv4_block4_2_bn",
          "118_Conv2D": "conv4_block4_3_conv",
          "119_BatchNormalization": "conv4_block4_3_bn",
          "122_Conv2D": "conv4_block5_1_conv",
          "123_BatchNormalization": "conv4_block5_1_bn",
          "125_Conv2D": "conv4_block5_2_conv",
          "126_BatchNormalization": "conv4_block5_2_bn",
          "128_Conv2D": "conv4_block5_3_conv",
          "129_BatchNormalization": "conv4_block5_3_bn",
          "132_Conv2D": "conv4_block6_1_conv",
          "133_BatchNormalization": "conv4_block6_1_bn",
          "135_Conv2D": "conv4_block6_2_conv",
          "136_BatchNormalization": "conv4_block6_2_bn",
          "138_Conv2D": "conv4_block6_3_conv",
          "139_BatchNormalization": "conv4_block6_3_bn",
          "142_Conv2D": "conv5_block1_1_conv",
          "143_BatchNormalization": "conv5_block1_1_bn",
          "145_Conv2D": "conv5_block1_2_conv",
          "146_BatchNormalization": "conv5_block1_2_bn",
          "148_Conv2D": "conv5_block1_3_conv",
          "149_BatchNormalization": "conv5_block1_3_bn",
          "150_Conv2D": "conv5_block1_0_conv",
          "151_BatchNormalization": "conv5_block1_0_bn",
          "154_Conv2D": "conv5_block2_1_conv",
          "155_BatchNormalization": "conv5_block2_1_bn",
          "157_Conv2D": "conv5_block2_2_conv",
          "158_BatchNormalization": "conv5_block2_2_bn",
          "160_Conv2D": "conv5_block2_3_conv",
          "161_BatchNormalization": "conv5_block2_3_bn",
          "164_Conv2D": "conv5_block3_1_conv",
          "165_BatchNormalization": "conv5_block3_1_bn",
          "167_Conv2D": "conv5_block3_2_conv",
          "168_BatchNormalization": "conv5_block3_2_bn",
          "170_Conv2D": "conv5_block3_3_conv",
          "171_BatchNormalization": "conv5_block3_3_bn",
          "175_FC": "probs"}

# PyDTNN <-> Keras weights name conversion
weights = {"kernel": Parameters.WEIGHTS,
           "bias": Parameters.BIASES,
           Parameters.GAMMA: Parameters.GAMMA,
           Parameters.BETA: Parameters.BETA,
           "moving_mean": "running_mean",
           "moving_variance": "running_var"}


# ResNet-50-v1.5 HDF5 file
IN_FILE = 'resnet50_weights_tf_dim_ordering_tf_kernels.h5'
OUT_FILE = 'resnet50_weights_pydtnn_kernels.npz'
PATH = 'https://storage.googleapis.com/tensorflow/keras-applications/resnet/'

if __name__ == "__main__":
    # Download weights file
    if not os.path.exists(IN_FILE):
        try:
            urllib.request.urlretrieve(PATH + IN_FILE, IN_FILE)
        except BaseException as exc:
            raise SystemExit(f"Error while downloading {PATH + IN_FILE}") from exc

    f = h5py.File("resnet50_weights_tf_dim_ordering_tf_kernels.h5", "r")
    out = {}
    for pydtnn_label, keras_label in layers.items():
        for weight_label in f[keras_label].attrs["weight_names"]:
            w = weight_label.decode("utf-8").split("/")[1].split(":")[0]
            w = weights[w]
            value = f[keras_label][weight_label][:]
            if value.ndim == 4:
                value = value.transpose(3, 2, 0, 1)
            out[f"{pydtnn_label}_{w}"] = value
    f.close()
    np.savez_compressed(OUT_FILE.split(".")[0], **out)

    print(f"Successfully written PyDTNN output file {OUT_FILE}!")
